""" Implements various models.

:author: Michel Bierlaire
:date: Fri Mar 29 17:13:14 2019
"""

import numpy as np
import biogeme.exceptions as excep

#from biogeme.expressions import Plus, Minus, Times, Divide, Power, bioMin, bioMax, And, Or, Equal, NotEqual, LessOrEqual, GreaterOrEqual, Less, Greater, UnaryMinus, MonteCarlo, bioNormalPdf, bioNormalCdf, PanelLikelihoodTrajectory, exp, log, Derive, Integrate, bioDraws, Numeric, Variable, DefineVariable, RandomVariable, Beta, bioLogLogit, bioLogLogitFullChoiceSet, bioMultSum, Elem, bioLinearUtility

from biogeme.expressions import _bioLogLogit, _bioLogLogitFullChoiceSet, exp, log, Elem, bioMin, bioMax, Numeric, Beta, bioMultSum, Variable

def loglogit(V,av,i):
    """The logarithm of the logit model 

    The model is defined as 
    
    .. math:: \\frac{a_i e^{V_i}}{\\sum_{i=1}^J a_j e^{V_j}}

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :type V: dict(int:biogeme.expressions.Expression)

    :param av: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type av: dict(int:biogeme.expressions.Expression)

    :param i: id of the alternative for which the probability must be
              calculated.
    :type i: int

    :return: choice probability of alternative number i.
    :rtype: biogeme.expressions.Expression
    """

    if av is None:
        return _bioLogLogitFullChoiceSet(V,av=None,choice=i)
    else:
        return _bioLogLogit(V,av,i)

def logit(V,av,i):
    """The logit model 

    The model is defined as 
    
    .. math:: \\frac{a_i e^{V_i}}{\\sum_{i=1}^J a_j e^{V_j}}

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :type V: dict(int:biogeme.expressions.Expression)

    :param av: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type av: dict(int:biogeme.expressions.Expression)

    :param i: id of the alternative for which the probability must be
              calculated.
    :type i: int

    :return: choice probability of alternative number i.
    :rtype: biogeme.expressions.Expression

    """
    if av is None:
        return exp(_bioLogLogitFullChoiceSet(V,av=None,choice=i))
    else:
        return exp(_bioLogLogit(V,av,i))


def boxcox(x,l):
    """Box-Cox transform 

    .. math:: B(x,\\ell) = \\frac{x^{\\ell}-1}{\\ell}. 

    It has the property that 

    .. math:: \\lim_{\\ell \\to 0} B(x,\\ell)=\\log(x).

    :param x: a variable to transform.
    :type x: biogeme.expressions.Expression
    :param l: parameter of the transformation.
    :type l: biogeme.expressions.Expression
    
    :return: the Box-Cox transform
    :rtype: biogeme.expressions.Expression
    """
    if type(l) is float or type(l) is int:
        if l < 0:
            raise excep.biogemeError(f"The parameter of a Box-Cox transform must be non negative, and not {l}.")
        if l < np.sqrt(np.finfo(np.float64).eps):
            if type(x) is float or type(x) is int:
                return np.log(x)
            else:
                return log(x)
        else:
            return (x**l-1.0)/l
    else:
        return Elem({0:(x**l-1.0)/l,1:log(x)},l < Numeric(np.sqrt(np.finfo(np.float64).eps)))
                

def piecewise(variable,thresholds):
    raise excep.biogemeError("The function 'piecewise' is obsolete and has been replaced by 'piecewiseVariables'. Its use has changed. Please refer to the documentation.")

def piecewiseVariables(variable,thresholds):
    """Generate the variables to include in a piecewise linear specification.

    If there are K thresholds, K-1 variables are generated. The first
    and last thresholds can be defined as None, corresponding to
    :math:`-\\infty` and :math:`+\infty`,respectively. If :math:`t` is
    the variable of interest, for each interval :math:`[a:a+b[`, we
    define a variable defined as:

    .. math:: x_{Ti} =\\left\\{  \\begin{array}{ll} 0 & \\text{if } t < a \\\\ t-a & \\text{if } a \\leq t < a+b \\\\ b  & \\text{otherwise}  \\end{array}\\right. \\;\\;\\;x_{Ti} = \\max(0,\\min(t-a,b)) 

    :param variable: variable for which we need the piecewise linear
       transform. The expression itself or the name of the variable can be given.
    :type variable: biogeme.expressions.Expression or str

    :param thresholds: list of thresholds
    :type thresholds: list(float)

    :return: list of variables to for the piecewise linear specification.
    :rtype: list(biogeme.expressions.Expression)

    .. seealso:: piecewiseFormula
    """
    I = len(thresholds)
    if all(t is None for t in thresholds):
        raise excep.biogemeError('All thresholds for the piecewise linear specification are set to None.')
    if None in thresholds[1:-1]:
        raise excep.biogemeError('For piecewise linear specification, only the first and the last thresholds can be None')

    # If the name of the variable is given, we transform it into an expression.
    if isinstance(variable,str):
        variable = Variable(variable)

    # First variable
    if thresholds[0] is None:
        results = [bioMin(variable,thresholds[1])]
    else:
        b = thresholds[1]-thresholds[0]
        results = [bioMax(Numeric(0),bioMin(variable - thresholds[0],b))]
    
    for i in range(1,I-2):
        b = thresholds[i+1]-thresholds[i]
        results += [bioMax(Numeric(0),bioMin(variable - thresholds[i],b))]

    # Last variable
    if thresholds[-1] is None:
        results += [bioMax(0,variable - thresholds[-2])]
    else:
        b = thresholds[-1]-thresholds[-2]
        results += [bioMax(Numeric(0),bioMin(variable - thresholds[-2],b))]
    return results

def piecewiseFormula(variable,thresholds,initialBetas = None):
    """Generate the formula for a piecewise linear specification.

    If there are K thresholds, K-1 variables are generated. The first
    and last thresholds can be defined as None, corresponding to
    :math:`-\\infty` and :math:`+\infty`,respectively. If :math:`t` is
    the variable of interest, for each interval :math:`[a:a+b[`, we
    define a variable defined as:
  
    .. math:: x_{Ti} =\\left\\{  \\begin{array}{ll} 0 & \\text{if } t < a \\\\ t-a & \\text{if } a \\leq t < a+b \\\\ b  & \\text{otherwise}  \\end{array}\\right. \\;\\;\\;x_{Ti} = \\max(0,\\min(t-a,b)) 

    New variables and new parameters are automatically created.

    :param variable: variable for which we need the piecewise linear
       transform.  
    :type string: name of the variable.

    :param thresholds: list of thresholds
    :type thresholds: list(float)

    :param initialBetas: list of values to initialize the beta
                         parameters.  The number of entries should be
                         the number of thresholds, plus one. If None,
                         the value of zero is used. Default: none.
    :type initialBetas: list(float)

    :return: expression of  the piecewise linear specification.
    :rtype: biogeme.expressions.Expression

    """

    I = len(thresholds)
    if all(t is None for t in thresholds):
        raise excep.biogemeError('All thresholds for the piecewise linear specification are set to None.')
    if None in thresholds[1:-1]:
        raise excep.biogemeError('For piecewise linear specification, only the first and the last thresholds can be None')
    if initialBetas is not None:
        if len(initialBetas) != I-1:
            raise excep.biogemeError(f'As there are {I} thresholds, a total of {I-1} values are needed to initialize the parameters. But {len(initialBetas)} are provided')
        
    vars = piecewiseVariables(Variable(f'{variable}'),thresholds)
    terms = []

    # First term
    betaValues = [0 if initialBetas is None else initialBetas[i] for i in range(I-1)]
    if thresholds[0] is None:
        beta = Beta(f'beta_{variable}_lessthan_{thresholds[1]}',betaValues[0],None,None,0)
    else:
        beta = Beta(f'beta_{variable}_{thresholds[0]}_{thresholds[1]}',betaValues[0],None,None,0)

    terms = [beta * vars[0]]

    # All terms, except the last
    for i in range(1,I-2):
        beta = Beta(f'beta_{variable}_{thresholds[i]}_{thresholds[i+1]}',betaValues[i],None,None,0)
        
        terms += [beta * vars[i]]

    # Last term
    if thresholds[-1] is None:
        beta = Beta(f'beta_{variable}_{thresholds[-2]}_more',betaValues[-2],None,None,0)
    else:
        beta = Beta(f'beta_{variable}_{thresholds[-2]}_{thresholds[-1]}',betaValues[-2],None,None,0)
    terms += [beta * vars[-1]]
    return bioMultSum(terms)

def piecewiseFunction(x,thresholds,betas):
    """Plot a piecewise linear specification.

    If there are K thresholds, K-1 variables are generated. The first
    and last thresholds can be defined as None, corresponding to
    :math:`-\\infty` and :math:`+\infty`,respectively. If :math:`t` is
    the variable of interest, for each interval :math:`[a:a+b[`, we
    define a variable defined as:
  
    .. math:: x_{Ti} =\\left\\{  \\begin{array}{ll} 0 & \\text{if } t < a \\\\ t-a & \\text{if } a \\leq t < a+b \\\\ b  & \\text{otherwise}  \\end{array}\\right. \\;\\;\\;x_{Ti} = \\max(0,\\min(t-a,b)) 

    :param x: value at which the piecewise specification must be avaluated
    :type x: float

    :param thresholds: list of thresholds
    :type thresholds: list(float)

    :param betas: list of the beta parameters.  The number of entries
                         should be the number of thresholds, plus
                         one.
    :type betas: list(float)

    :return: value of the numpy function
    :rtype: float
    """
    I = len(thresholds)
    if all(t is None for t in thresholds):
        raise excep.biogemeError('All thresholds for the piecewise linear specification are set to None.')
    if None in thresholds[1:-1]:
        raise excep.biogemeError('For piecewise linear specification, only the first and the last thresholds can be None')
    if len(betas) != I-1:
        raise excep.biogemeError(f'As there are {I} thresholds, a total of {I-1} values are needed to initialize the parameters. But {len(betas)} are provided')

    # If the first threshold is not -infinity, we need to check if
    # x is beyond it.
    if thresholds[0] is not None:
        if x < thresholds[0]:
            return 0
    rest = x
    total = 0
    for i in range(len(betas)):
        if thresholds[i+1] is None:
            total += betas[i] * rest
            return total
        elif x < thresholds[i+1]:
            total += betas[i] * rest
            return total
        else:
            total += betas[i] * (thresholds[i+1] - (0 if thresholds[i] is None else thresholds[i]))
            rest = x - thresholds[i+1]
    return total
    
    
def logmev(V,logGi,av,choice) :
    """ Log of the choice probability for a MEV model.
    
    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :type V: dict(int:biogeme.expressions.Expression)

    :param logGi: a dictionary mapping each alternative id with the function

    .. math:: \\ln \\frac{\\partial G}{\\partial y_i}(e^{V_1},\\ldots,e^{V_J})
        
    where :math:`G` is the MEV generating function. If an alternative :math:`i` is not available, then :math:`G_i = 0`.

    :type logGi: dict(int:biogeme.expressions.Expression)

    :param av: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type av: dict(int:biogeme.expressions.Expression)

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.Expression

    :return: log of the choice probability of the MEV model, given by
    
    .. math:: V_i + \\ln G_i(e^{V_1},\\ldots,e^{V_J}) - \\ln\\left(\\sum_j e^{V_j + \\ln G_j(e^{V_1},\\ldots,e^{V_J})}\\right)

    """
    H = {i:v + logGi[i] for i,v in V.items()}
    if av is None:
        logP = _bioLogLogitFullChoiceSet(H,av=None,choice=choice)
    else:
        logP = _bioLogLogit(H,av,choice)
    return logP

def mev(V,logGi,av,choice) :
    """ Choice probability for a MEV model.

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :type V: dict(int:biogeme.expressions.Expression)


    :param logGi: a dictionary mapping each alternative id with the function

    .. math:: \\ln \\frac{\\partial G}{\\partial y_i}(e^{V_1},\\ldots,e^{V_J})
        
    where :math:`G` is the MEV generating function. If an alternative :math:`i` is not available, then :math:`G_i = 0`.

    :param av: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type av: dict(int:biogeme.expressions.Expression)

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.Expression

    :return: Choice probability of the MEV model, given by

    .. math:: \\frac{e^{V_i + \\ln G_i(e^{V_1},\\ldots,e^{V_J})}}{\\sum_j e^{V_j + \\ln G_j(e^{V_1},\\ldots,e^{V_J})}}

    """
    return exp(logmev(V,logGi,av,choice))

def logmev_endogenousSampling(V,logGi,av,correction,choice) :
    """Log of choice probability for a MEV model, including the correction for endogenous sampling as proposed by `Bierlaire, Bolduc and McFadden (2008)`_.

    .. _`Bierlaire, Bolduc and McFadden (2008)`: http://dx.doi.org/10.1016/j.trb.2007.09.003

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.


    :param logGi: a dictionary mapping each alternative id with the function

    .. math:: \\ln \\frac{\\partial G}{\\partial y_i}(e^{V_1},\\ldots,e^{V_J})
        
    where :math:`G` is the MEV generating function. If an alternative :math:`i` is not available, then :math:`G_i = 0`.

    :type logGi: dict(int:biogeme.expressions.Expression)

    :param av: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type av: dict(int:biogeme.expressions.Expression)

    
    :param correction: a dict of expressions for the correstion terms
                       of each alternative.
    :type correction: dict(int:biogeme.expressions.Expression)

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.Expression

    :return: log of the choice probability of the MEV model, given by

    .. math:: V_i + \\ln G_i(e^{V_1},\\ldots,e^{V_J}) + \\omega_i - \\ln\\left(\\sum_j e^{V_j + \\ln G_j(e^{V_1},\\ldots,e^{V_J})+ \\omega_j}\\right)

    where :math:`\\omega_i` is the correction term for alternative :math:`i`.
    """
    H = {i: v + logGi[i] + correction[i] for i,v in V.items()}
    logP = _bioLogLogit(H,av,choice)
    return logP



def mev_endogenousSampling(V,logGi,av,correction,choice) :
    """Choice probability for a MEV model, including the correction for endogenous sampling as proposed by `Bierlaire, Bolduc and McFadden (2008)`_.

    .. _`Bierlaire, Bolduc and McFadden (2008)`: http://dx.doi.org/10.1016/j.trb.2007.09.003

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.


    :param logGi: a dictionary mapping each alternative id with the function

    .. math:: \\ln \\frac{\\partial G}{\\partial y_i}(e^{V_1},\\ldots,e^{V_J})
        
    where :math:`G` is the MEV generating function. If an alternative :math:`i` is not available, then :math:`G_i = 0`.

    :type logGi: dict(int:biogeme.expressions.Expression)

    :param av: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type av: dict(int:biogeme.expressions.Expression)

    
    :param correction: a dict of expressions for the correstion terms
                       of each alternative.
    :type correction: dict(int:biogeme.expressions.Expression)

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.Expression

    :return: log of the choice probability of the MEV model, given by

    .. math:: V_i + \\ln G_i(e^{V_1},\\ldots,e^{V_J}) + \\omega_i - \\ln\\left(\\sum_j e^{V_j + \\ln G_j(e^{V_1},\\ldots,e^{V_J})+ \\omega_j}\\right)

    where :math:`\\omega_i` is the correction term for alternative :math:`i`.
    """
    return exp(logmev_endogenousSampling(V,logGi,av,correction,choice))


def getMevForNested(V,availability,nests) :
    """ Implements the MEV generating function for the nested logit model

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.Expression)

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type availability: dict(int:biogeme.expressions.Expression)

    :param nests: A tuple containing as many items as nests. Each item is also a tuple containing two items

          - an object of type biogeme.expressions.Expression  representing the nest parameter,
          - a list containing the list of identifiers of the alternatives belonging to the nest.

      Example::

          nesta = MUA , [1,2,3]
          nestb = MUB , [4,5,6]
          nests = nesta, nestb


    
    :type nests: tuple

    :return: a dictionary mapping each alternative id with the function

    .. math:: \\ln \\frac{\\partial G}{\\partial y_i}(e^{V_1},\\ldots,e^{V_J}) = e^{(\\mu_m-1)V_i} \\left(\\sum_{i=1}^{J_m} e^{\\mu_m V_i}\\right)^{\\frac{1}{\\mu_m}-1}
    
    where :math:`m` is the (only) nest containing alternative :math:`i`, and :math:`G` is the MEV generating function.

    :rtype: dict(int:biogeme.expressions.Expression)

    """

    
    
    #y = {i:exp(v) for i,v in V.items()}
    logGi = {}
    for m in nests:
        if availability is None:
            sumdict = [exp(m[0] * V[i]) for i in m[1]]
        else:
            sumdict = [Elem({0:0.0,1: exp(m[0] * V[i])},availability[i]!=Numeric(0)) for i in m[1]]
        sum = bioMultSum(sumdict)
        for i in m[1]:
            logGi[i] = (m[0]-1.0) * V[i] + (1.0/m[0] - 1.0) * log(sum) 
    return logGi

def getMevForNestedMu(V,availability,nests,mu) :
    """Implements the MEV generating function for the nested logit model,
including the scale parameter

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.Expression)

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type availability: dict(int:biogeme.expressions.Expression)

    :param nests: A tuple containing as many items as nests. Each item is also a tuple containing two items

          - an object of type biogeme.expressions.Expression  representing the nest parameter,
          - a list containing the list of identifiers of the alternatives belonging to the nest.

      Example::

          nesta = MUA , [1,2,3]
          nestb = MUB , [4,5,6]
          nests = nesta, nestb

    :type nests: tuple

    :param mu: scale parameter
    :type mu: biogeme.expressions.Expression

    :return: a dictionary mapping each alternative id with the function

    .. math:: \\frac{\\partial G}{\\partial y_i}(e^{V_1},\\ldots,e^{V_J}) = \\mu e^{(\\mu_m-1)V_i} \\left(\\sum_{i=1}^{J_m} e^{\\mu_m V_i}\\right)^{\\frac{\\mu}{\\mu_m}-1}
    
    where :math:`m` is the (only) nest containing alternative :math:`i`, and :math:`G` is the MEV generating function.

    :rtype: dict(int:biogeme.expressions.Expression)

    """

    #y = {i:exp(v) for i,v in V.items()}
    logGi = {}
    for m in nests:
        if availability is None:
            sumdict = [exp(m[0] * V[i]) for i in m[1]]
        else:
            sumdict = [Elem({0:0.0,1: exp(m[0] * V[i])},availability[i]!=0) for i in m[1]]
        sum = bioMultSum(sumdict)
        for i in m[1]:
            logGi[i] = log(mu) + (m[0]-1.0) * V[i] + (mu/m[0] - 1.0) * log(sum) 
    return logGi




def nested(V,availability,nests,choice) :
    """Implements the nested logit model as a MEV model. 
    
    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.Expression)

    :param availability: dict of objects representing the availability
                         of each alternative, indexed by numerical
                         ids. Must be consistent with V, or None. In
                         this case, all alternatives are supposed to
                         be always available.

    :type availability: dict(int:biogeme.expressions.Expression)

    :param nests: A tuple containing as many items as nests. Each item is also a tuple containing two items
                  - an object of type biogeme.expressions.Expression  representing the nest parameter,
                  - a list containing the list of identifiers of the alternatives belonging to the nest.

      Example::

          nesta = MUA , [1,2,3]
          nestb = MUB , [4,5,6]
          nests = nesta, nestb
    
    :type nests: tuple

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.Expression

    :return: choice probability for the nested logit model,
             based on the derivatives of the MEV generating function produced
             by the function getMevForNested

    """

    ok, message = checkValidityNestedLogit(V,nests)
    if not ok:
        raise excep.biogemeError(message)
        
    logGi = getMevForNested(V,availability,nests)
    P = mev(V,logGi,availability,choice) 
    return P

def lognested(V,availability,nests,choice) :
    """Implements the log of a nested logit model as a MEV model. 
    
    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :type V: dict(int:biogeme.expressions.Expression)

    :param availability: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type availability: dict(int:biogeme.expressions.Expression)

    :param nests: A tuple containing as many items as nests. Each item is also a tuple containing two items

          - an object of type biogeme.expressions.Expression  representing the nest parameter,
          - a list containing the list of identifiers of the alternatives belonging to the nest.

      Example::

          nesta = MUA , [1,2,3]
          nestb = MUB , [4,5,6]
          nests = nesta, nestb
    
    :type nests: tuple

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.Expression

    :return: log of choice probability for the nested logit model,
             based on the derivatives of the MEV generating function produced
             by the function getMevForNested

    """
    ok, message = checkValidityNestedLogit(V,nests)
    if not ok:
        raise excep.biogemeError(message)
    logGi = getMevForNested(V,availability,nests)
    logP = logmev(V,logGi,availability,choice) 
    return logP

def nestedMevMu(V,availability,nests,choice,mu) :
    """Implements the nested logit model as a MEV model, where mu is also
    a parameter, if the user wants to test different normalization
    schemes.
 
    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :type V: dict(int:biogeme.expressions.Expression)

    :param availability: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type availability: dict(int:biogeme.expressions.Expression)

    :param nests: A tuple containing as many items as nests. Each item is also a tuple containing two items

          - an object of type biogeme.expressions.Expression  representing the nest parameter,
          - a list containing the list of identifiers of the alternatives belonging to the nest.

      Example::

          nesta = MUA , [1,2,3]
          nestb = MUB , [4,5,6]
          nests = nesta, nestb
    
    :type nests: tuple

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.Expression

    :param mu: expression producing the value of the top-level scale parameter.
    :type mu:  biogeme.expressions.Expression

    :return: the nested logit choice probability based on the following derivatives of the MEV generating function: 

    .. math:: \\frac{\\partial G}{\\partial y_i}(e^{V_1},\\ldots,e^{V_J}) = \\mu e^{(\\mu_m-1)V_i} \\left(\\sum_{i=1}^{J_m} e^{\\mu_m V_i}\\right)^{\\frac{\\mu}{\\mu_m}-1}

    Where :math:`m` is the (only) nest containing alternative :math:`i`, and
    :math:`G` is the MEV generating function.

    :rtype: biogeme.expressions.Expression

    """
    return exp(lognestedMevMu(V,availability,nests,choice,mu))

def lognestedMevMu(V,availability,nests,choice,mu) :
    """ Implements the log of the nested logit model as a MEV model, where mu is also a parameter, if the user wants to test different normalization schemes.


    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :type V: dict(int:biogeme.expressions.Expression)

    :param availability: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type availability: dict(int:biogeme.expressions.Expression)

    :param nests: A tuple containing as many items as nests. Each item is also a tuple containing two items

          - an object of type biogeme.expressions.Expression  representing the nest parameter,
          - a list containing the list of identifiers of the alternatives belonging to the nest.

      Example::

          nesta = MUA , [1,2,3]
          nestb = MUB , [4,5,6]
          nests = nesta, nestb
    
    :type nests: tuple

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.Expression

    :param mu: expression producing the value of the top-level scale parameter.
    :type mu:  biogeme.expressions.Expression

    :return: the log of the nested logit choice probability based on the following derivatives of the MEV generating function: 

    .. math:: \\frac{\\partial G}{\\partial y_i}(e^{V_1},\\ldots,e^{V_J}) = \\mu e^{(\\mu_m-1)V_i} \\left(\\sum_{i=1}^{J_m} e^{\\mu_m V_i}\\right)^{\\frac{\\mu}{\\mu_m}-1}

    where :math:`m` is the (only) nest containing alternative :math:`i`, and
    :math:`G` is the MEV generating function.

    :rtype: biogeme.expressions.Expression
    """
    
    y = {i:exp(v) for i,v in V.items()}
    logGi = getMevForNestedMu(V,availability,nests,mu)
    logP = logmev(V,logGi,availability,choice) 
    return logP

def cnl_avail(V,availability,nests,choice):
    """ Same as cnl. Maintained for backward compatibility

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.Expression)

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type availability: dict(int:biogeme.expressions.Expression)

    :param nests: a tuple containing as many items as nests. Each item is also a tuple containing two items

          - an object of type biogeme.expressions.Expression  representing the nest parameter,
          - a dictionary mapping the alternative ids with the cross-nested parameters for the corresponding nest. 

        Example::

            alphaA = {1: alpha1a,2: alpha2a, 3: alpha3a, 4: alpha4a,5: alpha5a, 6: alpha6a}
            alphaB = {1: alpha1b,2: alpha2b, 3: alpha3b, 4: alpha4b,5: alpha5b, 6: alpha6b}
            nesta = MUA , alphaA
            nestb = MUB , alphaB
            nests = nesta, nestb

    :type nests: tuple

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.Expression

    :return: choice probability for the cross-nested logit model.
    :rtype: biogeme.expressions.Expression
    """
    return cnl(V,availability,nests,choice)

def cnl(V,availability,nests,choice):
    """ Implements the cross-nested logit model as a MEV model. 

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.Expression)

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type available: dict(int:biogeme.expressions.Expression)

    :param nests: a tuple containing as many items as nests. Each item is also a tuple containing two items

          - an object of type biogeme.expressions.Expression  representing the nest parameter,
          - a dictionary mapping the alternative ids with the cross-nested parameters for the corresponding nest. 

        Example::

            alphaA = {1: alpha1a,2: alpha2a, 3: alpha3a, 4: alpha4a,5: alpha5a, 6: alpha6a}
            alphaB = {1: alpha1b,2: alpha2b, 3: alpha3b, 4: alpha4b,5: alpha5b, 6: alpha6b}
            nesta = MUA , alphaA
            nestb = MUB , alphaB
            nests = nesta, nestb

    :type nests: tuple

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.Expression

    :return: choice probability for the cross-nested logit model.
    :rtype: biogeme.expressions.Expression
    

    """
    return exp(logcnl(V,availability,nests,choice))

def logcnl_avail(V,availability,nests,choice) :
    """ Same as logcnl. Maintained for backward compatibility

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.Expression)

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type availability: dict(int:biogeme.expressions.Expression)

    :param nests: a tuple containing as many items as nests. Each item is also a tuple containing two items

          - an object of type biogeme.expressions.Expression  representing the nest parameter,
          - a dictionary mapping the alternative ids with the cross-nested parameters for the corresponding nest. 

        Example::

            alphaA = {1: alpha1a,2: alpha2a, 3: alpha3a, 4: alpha4a,5: alpha5a, 6: alpha6a}
            alphaB = {1: alpha1b,2: alpha2b, 3: alpha3b, 4: alpha4b,5: alpha5b, 6: alpha6b}
            nesta = MUA , alphaA
            nestb = MUB , alphaB
            nests = nesta, nestb

    :type nests: tuple

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.Expression

    :return: log of choice probability for the cross-nested logit model.
    :rtype: biogeme.expressions.Expression
    """
    return logcnl(V,availability,nests,choice)

def getMevForCrossNested(V,availability,nests) :
    """ Implements the MEV generating function for the cross-nested logit model as a MEV model. 
    
    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.Expression)

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type availability: dict(int:biogeme.expressions.Expression)

    :param nests: a tuple containing as many items as nests. Each item is also a tuple containing two items

          - an object of type biogeme.expressions.Expression  representing the nest parameter,
          - a dictionary mapping the alternative ids with the cross-nested parameters for the corresponding nest. 

        Example::

            alphaA = {1: alpha1a,2: alpha2a, 3: alpha3a, 4: alpha4a,5: alpha5a, 6: alpha6a}
            alphaB = {1: alpha1b,2: alpha2b, 3: alpha3b, 4: alpha4b,5: alpha5b, 6: alpha6b}
            nesta = MUA , alphaA
            nestb = MUB , alphaB
            nests = nesta, nestb

    :type nests: tuple

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.Expression

    :return: log of the choice probability for the cross-nested logit model.
    :rtype: biogeme.expressions.Expression

    """
    Gi_terms = {}
    logGi = {}
    for i in V:
        Gi_terms[i] = list()
    biosum = {}
    for m in nests:
        if availability is None:
            biosum = bioMultSum([a**(m[0]) * exp(m[0] * (V[i])) for i,a in m[1].items()])
        else:
            biosum = bioMultSum([availability[i] * a**(m[0]) * exp(m[0] * (V[i])) for i,a in m[1].items()])
        for i,a in m[1].items():
            Gi_terms[i] += [a**(m[0])* exp((m[0]-1) * (V[i])) * biosum **((1.0/m[0])-1.0)]  
    for k in V:
        logGi[k] = log(bioMultSum(Gi_terms[k]))
    return logGi

def logcnl(V,availability,nests,choice) :
    """ Implements the log of the cross-nested logit model as a MEV model. 
    
    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.Expression)

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type availability: dict(int:biogeme.expressions.Expression)

    :param nests: a tuple containing as many items as nests. Each item is also a tuple containing two items

          - an object of type biogeme.expressions.Expression  representing the nest parameter,
          - a dictionary mapping the alternative ids with the cross-nested parameters for the corresponding nest. 

        Example::

            alphaA = {1: alpha1a,2: alpha2a, 3: alpha3a, 4: alpha4a,5: alpha5a, 6: alpha6a}
            alphaB = {1: alpha1b,2: alpha2b, 3: alpha3b, 4: alpha4b,5: alpha5b, 6: alpha6b}
            nesta = MUA , alphaA
            nestb = MUB , alphaB
            nests = nesta, nestb

    :type nests: tuple

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.Expression

    :return: log of the choice probability for the cross-nested logit model.
    :rtype: biogeme.expressions.Expression

    """
    logGi = getMevForCrossNested(V,availability,nests)        
    logP = logmev(V,logGi,availability,choice) 
    return logP


def cnlmu(V,availability,nests,choice,mu) :
    """ Implements the cross-nested logit model as a MEV model with the homogeneity parameters is explicitly involved

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.Expression)

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type availability: dict(int:biogeme.expressions.Expression)

    :param nests: a tuple containing as many items as nests. Each item is also a tuple containing two items

          - an object of type biogeme.expressions.Expression  representing the nest parameter,
          - a dictionary mapping the alternative ids with the cross-nested parameters for the corresponding nest. 

        Example::

            alphaA = {1: alpha1a,2: alpha2a, 3: alpha3a, 4: alpha4a,5: alpha5a, 6: alpha6a}
            alphaB = {1: alpha1b,2: alpha2b, 3: alpha3b, 4: alpha4b,5: alpha5b, 6: alpha6b}
            nesta = MUA , alphaA
            nestb = MUB , alphaB
            nests = nesta, nestb

    :type nests: tuple

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.Expression

    :param mu: Homogeneity parameter :math:`\\mu`.
    :type mu: biogeme.expressions.Expression
    
    :return: choice probability for the cross-nested logit model.
    :rtype: biogeme.expressions.Expression
    """
    return exp(logcnlmu(V,availability,nests,choice,mu))

def getMevForCrossNestedMu(V,availability,nests,mu) :
    """ Implements the MEV generating function for the cross-nested logit model as a MEV model with the homogeneity parameters is explicitly involved.


    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.Expression)

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type availability: dict(int:biogeme.expressions.Expression)

    :param nests: a tuple containing as many items as nests. Each item is also a tuple containing two items

          - an object of type biogeme.expressions.Expression  representing the nest parameter,
          - a dictionary mapping the alternative ids with the cross-nested parameters for the corresponding nest. 

        Example::

            alphaA = {1: alpha1a,2: alpha2a, 3: alpha3a, 4: alpha4a,5: alpha5a, 6: alpha6a}
            alphaB = {1: alpha1b,2: alpha2b, 3: alpha3b, 4: alpha4b,5: alpha5b, 6: alpha6b}
            nesta = MUA , alphaA
            nestb = MUB , alphaB
            nests = nesta, nestb

    :type nests: tuple

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.Expression

    :param mu: Homogeneity parameter :math:`\\mu`.
    :type mu: biogeme.expressions.Expression
    
    :return: log of the choice probability for the cross-nested logit model.
    :rtype: biogeme.expressions.Expression

    """
    Gi_terms = {}
    logGi = {}
    for i in V:
        Gi_terms[i] = list()
    biosum = {}
    for m in nests:
        if availability is None:
            biosum = bioMultSum([a**(m[0]/mu) * exp(m[0] * (V[i])) for i,a in m[1].items()])
        else:
            biosum = bioMultSum([availability[i] * a**(m[0]/mu) * exp(m[0] * (V[i])) for i,a in m[1].items()])
        for i,a in m[1].items():
            Gi_terms[i] += [a**(m[0]/mu)* exp((m[0]-1) * (V[i])) * biosum **((mu/m[0])-1.0)]  
    for k in V:
        logGi[k] = log(mu * bioMultSum(Gi_terms[k]))
    return logGi

def logcnlmu(V,availability,nests,choice,mu):
    """ Implements the log of the cross-nested logit model as a MEV model with the homogeneity parameters is explicitly involved.


    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.Expression)

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with V, or
               None. In this case, all alternatives are supposed to be
               always available.

    :type availability: dict(int:biogeme.expressions.Expression)

    :param nests: a tuple containing as many items as nests. Each item is also a tuple containing two items

          - an object of type biogeme.expressions.Expression  representing the nest parameter,
          - a dictionary mapping the alternative ids with the cross-nested parameters for the corresponding nest. 

        Example::

            alphaA = {1: alpha1a,2: alpha2a, 3: alpha3a, 4: alpha4a,5: alpha5a, 6: alpha6a}
            alphaB = {1: alpha1b,2: alpha2b, 3: alpha3b, 4: alpha4b,5: alpha5b, 6: alpha6b}
            nesta = MUA , alphaA
            nestb = MUB , alphaB
            nests = nesta, nestb

    :type nests: tuple

    :param choice: id of the alternative for which the probability must be
              calculated.
    :type choice: biogeme.expressions.Expression

    :param mu: Homogeneity parameter :math:`\\mu`.
    :type mu: biogeme.expressions.Expression
    
    :return: log of the choice probability for the cross-nested logit model.
    :rtype: biogeme.expressions.Expression

    """
    logGi = getMevForCrossNestedMu(V,availability,nests,mu)
    logP = logmev(V,logGi,availability,choice) 
    return logP

def checkValidityNestedLogit(V,nests):
    """Verifies if the nested logit model is indeed based on a partition
       of the choice set.

    :param V: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.
    :type V: dict(int:biogeme.expressions.Expression)
    :param nests: A tuple containing as many items as nests. Each item is also a tuple containing two items

          - an object of type biogeme.expressions.Expression  representing the nest parameter,
          - a list containing the list of identifiers of the alternatives belonging to the nest.

      Example::

          nesta = MUA , [1,2,3]
          nestb = MUB , [4,5,6]
          nests = nesta, nestb

    :type nests: tuple


    :return: a tuple ok,message, where the message explains the
             problem is the nested structure is not OK.
    :rtype: tuple(bool, str)
    """

    ok = True
    message = 'The nested logit model is based on a partition. '
    
    fullChoiceSet = set([i for i,v in V.items()])
    unionOfNests = set.union(*[set(n[1]) for n in nests])
    if fullChoiceSet != unionOfNests:
        ok = False
        d1 = fullChoiceSet.difference(unionOfNests)
        d2 = unionOfNests.difference(fullChoiceSet)
        if d1:
            message += f'Alternatives in the choice set, but not in any nest: {d1}\n'
        if d2:
            message += f'Alternatives in a nest, but not in the choice set: {d2}\n'
    
    # Consider all pairs of nests and verify that the intersection is empty

    allPairs = [(n1, n2) for n1 in nests for n2 in nests if n1 != n2]
    for (n1,n2) in allPairs:
        inter = set(n1[1]).intersection(n2[1])
        if inter:
            ok = False
            message += f'Two nests contain the following alternative(s): {inter}\n'
    return ok, message




"""File 02nestedSimulation.py

:author: Michel Bierlaire, EPFL
:date: Wed Sep 11 10:23:51 2019

 We use a previously estimated nested logit model.
 Three alternatives: public transporation, car and slow modes.
 RP data.
 We simulate market shares and revenues. 
"""
import sys
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.models as models
import biogeme.results as res
from biogeme.expressions import Beta, DefineVariable

print("Running 02nestedSimulation.py...")

# Read the data
df = pd.read_csv("optima.dat",sep='\t')
database = db.Database("optima",df)

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
#print(database.data.describe())

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Exclude observations such that the chosen alternative is -1
exclude = (Choice == -1.0)
database.remove(exclude)

# Normalize the weights
sumWeight = database.data['Weight'].sum()
normalizedWeight = Weight * 1906 / 0.814484

# Calculate the number of accurences of a value in the database
numberOfMales = database.count("Gender",1)
print(f"Number of males:   {numberOfMales}")
numberOfFemales = database.count("Gender",2)
print(f"Number of females: {numberOfFemales}")

# For more complex conditions, we use Pandas
unreportedGender = \
                   database.data[(database.data["Gender"] != 1)
                    & (database.data["Gender"] != 2)].count()["Gender"]
print(f"Unreported gender: {unreportedGender}")

# List of parameters. Their value will be set later.
ASC_CAR = Beta('ASC_CAR',0,None,None,0)
ASC_PT = Beta('ASC_PT',0,None,None,1)
ASC_SM = Beta('ASC_SM',0,None,None,0)
BETA_TIME_FULLTIME = Beta('BETA_TIME_FULLTIME',0,None,None,0)
BETA_TIME_OTHER = Beta('BETA_TIME_OTHER',0,None,None,0)
BETA_DIST_MALE = Beta('BETA_DIST_MALE',0,None,None,0)
BETA_DIST_FEMALE = Beta('BETA_DIST_FEMALE',0,None,None,0)
BETA_DIST_UNREPORTED = Beta('BETA_DIST_UNREPORTED',0,None,None,0)
BETA_COST = Beta('BETA_COST',0,None,None,0)

# Definition of variables:
# For numerical reasons, it is good practice to scale the data to
# that the values of the parameters are around 1.0.

# The following statements are designed to preprocess the data.
# It is like creating a new columns in the data file. This
# should be preferred to the statement like
# TimePT_scaled = Time_PT / 200.0
# which will cause the division to be reevaluated again and again,
# throuh the iterations. For models taking a long time to
# estimate, it may make a significant difference.

TimePT_scaled = TimePT / 200
TimeCar_scaled = TimeCar / 200
MarginalCostPT_scaled = MarginalCostPT / 10
CostCarCHF_scaled = CostCarCHF / 10
distance_km_scaled = distance_km / 5

male = (Gender == 1)
female = (Gender == 2)
unreportedGender = (Gender == -1)

fulltime = (OccupStat == 1)
notfulltime = (OccupStat != 1)

# Definition of utility functions:
V_PT = ASC_PT + BETA_TIME_FULLTIME * TimePT_scaled * fulltime + \
       BETA_TIME_OTHER * TimePT_scaled * notfulltime + \
       BETA_COST * MarginalCostPT_scaled
V_CAR = ASC_CAR + \
        BETA_TIME_FULLTIME * TimeCar_scaled * fulltime + \
        BETA_TIME_OTHER * TimeCar_scaled * notfulltime + \
        BETA_COST * CostCarCHF_scaled
V_SM = ASC_SM + \
       BETA_DIST_MALE * distance_km_scaled * male + \
       BETA_DIST_FEMALE * distance_km_scaled * female + \
       BETA_DIST_UNREPORTED * distance_km_scaled * unreportedGender

# Associate utility functions with the numbering of alternatives
V = {0: V_PT,
     1: V_CAR,
     2: V_SM}

# Associate the availability conditions with the alternatives.
# In this example all alternatives are available for each individual.
av = {0: 1,
      1: 1,
      2: 1}

# Definition of the nests:
# 1: nests parameter
# 2: list of alternatives

MU_NOCAR = Beta('MU_NOCAR',1.0,1.0,None,0)

CAR_NEST = 1.0 , [ 1]
NO_CAR_NEST = MU_NOCAR , [ 0, 2]
nests = CAR_NEST, NO_CAR_NEST

# The choice model is a nested logit
prob_pt = models.nested(V,av,nests,0)
prob_car = models.nested(V,av,nests,1)
prob_sm = models.nested(V,av,nests,2)

simulate = {'weight': normalizedWeight,
            'Prob. car': prob_car,
            'Prob. public transportation': prob_pt,
            'Prob. slow modes':prob_sm,
            'Revenue public transportation':prob_pt * MarginalCostPT}

biogeme  = bio.BIOGEME(database,simulate)
biogeme.modelName = "02nestedSimulation"

# Retrieve the names of the parameters
betas = biogeme.freeBetaNames

# Read the estimation results from the file
results = res.bioResults(pickleFile='01nestedEstimation.pickle')

# Extract the values that are necessary
betaValues = results.getBetaValues()

# simulatedValues is a Panda dataframe with the same number of rows as
# the database, and as many columns as formulas to simulate.
simulatedValues = biogeme.simulate(betaValues)

# Calculate confidence intervals
b = results.getBetasForSensitivityAnalysis(betas,size=100)

# Returns data frame containing, for each simulated value, the left
# and right bounds of the confidence interval calculated by
# simulation.
left,right = biogeme.confidenceIntervals(b,0.9)

# We calculate now the market shares and their confidence intervals

simulatedValues['Weighted prob. car'] = \
  simulatedValues['weight'] * simulatedValues['Prob. car']
left['Weighted prob. car'] = left['weight'] * left['Prob. car']
right['Weighted prob. car'] = right['weight'] * right['Prob. car']

marketShare_car = simulatedValues['Weighted prob. car'].mean()
marketShare_car_left = left['Weighted prob. car'].mean()
marketShare_car_right = right['Weighted prob. car'].mean()
print(f"Market share for car: {100*marketShare_car:.1f}% [{100*marketShare_car_left:.1f}%,{100*marketShare_car_right:.1f}%]")

simulatedValues['Weighted prob. PT'] = simulatedValues['weight'] * simulatedValues['Prob. public transportation']
marketShare_pt = simulatedValues['Weighted prob. PT'].mean()
marketShare_pt_left = (left['Prob. public transportation'] * left['weight']).mean()
marketShare_pt_right = (right['Prob. public transportation'] * right['weight']).mean()
print(f"Market share for PT: {100*marketShare_pt:.1f}% [{100*marketShare_pt_left:.1f}%,{100*marketShare_pt_right:.1f}%]")

marketShare_sm = (simulatedValues['Prob. slow modes'] *
                  simulatedValues['weight']).mean()
marketShare_sm_left = (left['Prob. slow modes'] * left['weight']).mean()
marketShare_sm_right = (right['Prob. slow modes'] * right['weight']).mean()
print(f"Market share for slow modes: {100*marketShare_sm:.1f}% [{100*marketShare_sm_left:.1f}%,{100*marketShare_sm_right:.1f}%]")

# and, similarly, the revenues
revenues_pt = (simulatedValues['Revenue public transportation'] *
               simulatedValues['weight']).sum()
revenues_pt_left = (left['Revenue public transportation'] *
                    left['weight']).sum()
revenues_pt_right = (right['Revenue public transportation'] *
                     right['weight']).sum()
print(f"Revenues for PT: {revenues_pt:.3f} [{revenues_pt_left:.3f},{revenues_pt_right:.3f}]")



"""File 01nestedEstimation.py

:author: Michel Bierlaire, EPFL
:date: Wed Sep 11 09:59:55 2019

 Estimation of a nested logit model, that will be used for simuation.
 Three alternatives: public transporation, car and slow modes.
 RP data.
"""
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.models as models
from biogeme.expressions import Beta, DefineVariable

# Read the data
df = pd.read_csv("optima.dat",sep='\t')
database = db.Database("optima",df)

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
#print(database.data.describe())

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Exclude observations such that the chosen alternative is -1
exclude = (Choice == -1.0)
database.remove(exclude)


# List of parameters to be estimated
ASC_CAR = Beta('ASC_CAR',0,None,None,0)
ASC_PT = Beta('ASC_PT',0,None,None,1)
ASC_SM = Beta('ASC_SM',0,None,None,0)
BETA_TIME_FULLTIME = Beta('BETA_TIME_FULLTIME',0,None,None,0)
BETA_TIME_OTHER = Beta('BETA_TIME_OTHER',0,None,None,0)
BETA_DIST_MALE = Beta('BETA_DIST_MALE',0,None,None,0)
BETA_DIST_FEMALE = Beta('BETA_DIST_FEMALE',0,None,None,0)
BETA_DIST_UNREPORTED = Beta('BETA_DIST_UNREPORTED',0,None,None,0)
BETA_COST = Beta('BETA_COST',0,None,None,0)


# Definition of variables:
# For numerical reasons, it is good practice to scale the data to
# that the values of the parameters are around 1.0.

# The following statements are designed to preprocess the data.
# It is like creating a new columns in the data file. This
# should be preferred to the statement like
# TimePT_scaled = Time_PT / 200.0
# which will cause the division to be reevaluated again and again,
# throuh the iterations. For models taking a long time to
# estimate, it may make a significant difference.

TimePT_scaled = TimePT / 200
TimeCar_scaled = TimeCar / 200
MarginalCostPT_scaled = MarginalCostPT / 10 
CostCarCHF_scaled = CostCarCHF / 10
distance_km_scaled = distance_km / 5
male = (Gender == 1)
female = (Gender == 2)
unreportedGender = (Gender == -1)

fulltime = (OccupStat == 1)
notfulltime = (OccupStat != 1)

# Definition of utility functions:
V_PT = ASC_PT + BETA_TIME_FULLTIME * TimePT_scaled * fulltime + \
       BETA_TIME_OTHER * TimePT_scaled * notfulltime + \
       BETA_COST * MarginalCostPT_scaled
V_CAR = ASC_CAR + \
        BETA_TIME_FULLTIME * TimeCar_scaled * fulltime + \
        BETA_TIME_OTHER * TimeCar_scaled * notfulltime + \
        BETA_COST * CostCarCHF_scaled
V_SM = ASC_SM + \
       BETA_DIST_MALE * distance_km_scaled * male + \
       BETA_DIST_FEMALE * distance_km_scaled * female + \
       BETA_DIST_UNREPORTED * distance_km_scaled * unreportedGender

# Associate utility functions with the numbering of alternatives
V = {0: V_PT,
     1: V_CAR,
     2: V_SM}

# Associate the availability conditions with the alternatives.
# In this example all alternatives are available for each individual.
av = {0: 1,
      1: 1,
      2: 1}

# Definition of the nests:
# 1: nests parameter
# 2: list of alternatives

MU_NOCAR = Beta('MU_NOCAR',1.0,1.0,None,0)

CAR_NEST = 1.0 , [ 1]
NO_CAR_NEST = MU_NOCAR , [ 0, 2]
nests = CAR_NEST, NO_CAR_NEST

# The choice model is a nested logit, with availability conditions
logprob = models.lognested(V,av,nests,Choice)

# Define level of verbosity
import biogeme.messaging as msg
logger = msg.bioMessage()
logger.setSilent()
#logger.setWarning()
#logger.setGeneral()
#logger.setDetailed()

# Create the Biogeme object
biogeme  = bio.BIOGEME(database,logprob)
biogeme.modelName = "01nestedEstimation"

# Estimate the parameters. Calculate also the standard errors using
# bootstrapping.
results = biogeme.estimate(bootstrap=100)

print("Estimated betas: {}".format(len(results.data.betaValues)))





""" File: 04normalMixtureNumerical.py

 Author: Michel Bierlaire, EPFL
 Date: Wed Dec 11 17:06:52 2019

Calculation of a mixtures of logit models where the integral is
calculated using numerical integration.

"""

import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.draws as draws
import biogeme.distributions as dist
import biogeme.models as models

from biogeme.expressions import Beta, DefineVariable, RandomVariable, Integrate

p = pd.read_csv("swissmetro.dat",sep='\t')
# Use only the first observation (index 0)
p = p.drop(p[p.index != 0].index)
database = db.Database("swissmetro",p)

globals().update(database.variables)

#Parameters 
ASC_CAR = 0.137
ASC_TRAIN = -0.402
ASC_SM = 0
B_TIME = -2.26
B_TIME_S = 1.66
B_COST = -1.29

# Define a random parameter, normally distributed, 
# designed to be used for integration
omega = RandomVariable('omega')
density = dist.normalpdf(omega) 
B_TIME_RND = B_TIME + B_TIME_S * omega

# Utility functions

#If the person has a GA (season ticket) her 
#incremental cost is actually 0 
#rather than the cost value gathered from the
# network data. 
SM_COST =  SM_CO   * (  GA   ==  0  ) 
TRAIN_COST =  TRAIN_CO   * (  GA   ==  0  )

# For numerical reasons, it is good practice to scale the data to
# that the values of the parameters are around 1.0. 
# A previous estimation with the unscaled data has generated
# parameters around -0.01 for both cost and time. 
# Therefore, time and cost are multipled my 0.01.

TRAIN_TT_SCALED = \
  DefineVariable('TRAIN_TT_SCALED', TRAIN_TT / 100.0,database)
TRAIN_COST_SCALED = \
  DefineVariable('TRAIN_COST_SCALED', TRAIN_COST / 100,database)
SM_TT_SCALED = DefineVariable('SM_TT_SCALED', SM_TT / 100.0,database)
SM_COST_SCALED = DefineVariable('SM_COST_SCALED', SM_COST / 100,database)
CAR_TT_SCALED = DefineVariable('CAR_TT_SCALED', CAR_TT / 100,database)
CAR_CO_SCALED = DefineVariable('CAR_CO_SCALED', CAR_CO / 100,database)
CAR_AV_SP =  DefineVariable('CAR_AV_SP',CAR_AV  * (  SP   !=  0  ),database)
TRAIN_AV_SP =  DefineVariable('TRAIN_AV_SP',TRAIN_AV  * (  SP   !=  0  ),database)

V1 = ASC_TRAIN + \
     B_TIME_RND * TRAIN_TT_SCALED + \
     B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + \
     B_TIME_RND * SM_TT_SCALED + \
     B_COST * SM_COST_SCALED
V3 = ASC_CAR + \
     B_TIME_RND * CAR_TT_SCALED + \
     B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1,
     2: V2,
     3: V3}

# Associate the availability conditions with the alternatives

av = {1: TRAIN_AV_SP,
      2: SM_AV,
      3: CAR_AV_SP}

# The choice model is a logit, with availability conditions
integrand = models.logit(V,av,CHOICE)
numericalI = Integrate(integrand*density,'omega')

simulate = {'Numerical': numericalI}

biogeme = bio.BIOGEME(database,simulate)
results = biogeme.simulate()
print('Mixture of logit - numerical integration: ',results.iloc[0]['Numerical'])



"""File 01logitBis.py

:author: Michel Bierlaire, EPFL
:date: Thu Sep  6 15:14:39 2018

 Logit model
 Three alternatives: Train, Car and Swissmetro
 SP data
"""

import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.optimization as opt
import biogeme.models as models
from biogeme.expressions import Beta, DefineVariable,bioLinearUtility

# Read the data
df = pd.read_csv("swissmetro.dat",'\t')
database = db.Database("swissmetro",df)

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
#print(database.data.describe())

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Removing some observations can be done directly using pandas.
#remove = (((database.data.PURPOSE != 1) & (database.data.PURPOSE != 3)) | (database.data.CHOICE == 0))
#database.data.drop(database.data[remove].index,inplace=True)

# Here we use the "biogeme" way for backward compatibility
exclude = (( PURPOSE != 1 ) * (  PURPOSE   !=  3  ) +  ( CHOICE == 0 )) > 0
database.remove(exclude)


# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR',0,None,None,0)
ASC_TRAIN = Beta('ASC_TRAIN',0,None,None,0)
ASC_SM = Beta('ASC_SM',0,None,None,1)
B_TIME = Beta('B_TIME',0,None,None,0)
B_COST = Beta('B_COST',0,None,None,0)

# Definition of new variables
SM_COST =  SM_CO   * (  GA   ==  0  ) 
TRAIN_COST =  TRAIN_CO   * (  GA   ==  0  )

# Definition of new variables: adding columns to the database 
CAR_AV_SP =  DefineVariable('CAR_AV_SP',CAR_AV  * (  SP   !=  0  ),database)
TRAIN_AV_SP =  DefineVariable('TRAIN_AV_SP',TRAIN_AV  * (  SP   !=  0  ),database)
TRAIN_TT_SCALED = DefineVariable('TRAIN_TT_SCALED',\
                                 TRAIN_TT / 100.0,database)
TRAIN_COST_SCALED = DefineVariable('TRAIN_COST_SCALED',\
                                   TRAIN_COST / 100,database)
SM_TT_SCALED = DefineVariable('SM_TT_SCALED', SM_TT / 100.0,database)
SM_COST_SCALED = DefineVariable('SM_COST_SCALED', SM_COST / 100,database)
CAR_TT_SCALED = DefineVariable('CAR_TT_SCALED', CAR_TT / 100,database)
CAR_CO_SCALED = DefineVariable('CAR_CO_SCALED', CAR_CO / 100,database)

# We explicitly mention that the utility functions are linear.
# They are specified as a list of terms.
terms1 = [(B_TIME,TRAIN_TT_SCALED),(B_COST,TRAIN_COST_SCALED)] 
V1 = ASC_TRAIN + bioLinearUtility(terms1)

terms2 = [(B_TIME,SM_TT_SCALED),(B_COST,SM_COST_SCALED)] 
V2 = ASC_SM + bioLinearUtility(terms2)

terms3 = [(B_TIME,CAR_TT_SCALED),(B_COST,CAR_CO_SCALED)] 
V3 = ASC_CAR + bioLinearUtility(terms3)

# Associate utility functions with the numbering of alternatives
V = {1: V1,
     2: V2,
     3: V3}

# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP,
      2: SM_AV,
      3: CAR_AV_SP}

# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob = models.loglogit(V,av,CHOICE)

# Define level of verbosity
import biogeme.messaging as msg
logger = msg.bioMessage()
logger.setSilent()
#logger.setWarning()
#logger.setGeneral()
#logger.setDetailed()

# Create the Biogeme object
biogeme  = bio.BIOGEME(database,logprob)
biogeme.modelName = "01logitBis"

# Avoids the generation of the pickle file.  Note: this is included
# only in this example to illustrate the syntax, and not recommended.
# The pickle file contains the results of the estimation and allows
# you to analyze the results later on without re-estimating the model.
biogeme.generatePickle = False

# Estimate the parameters. 
results = biogeme.estimate(algorithm=opt.newtonTrustRegionForBiogeme,bootstrap=100)

# Print the estimated values
betas = results.getBetaValues()
for k,v in betas.items():
    print(f"{k:10}=\t{v:.3g}")

# Get the results in a pandas table
pandasResults = results.getEstimatedParameters()
print(pandasResults)



import pandas as pd

import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.models as models
from biogeme.expressions import Beta, DefineVariable, Derive
import unittest

pandas = pd.read_csv("swissmetro.dat",sep='\t')
database = db.Database("swissmetro",pandas)

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
#print(database.data.describe())

globals().update(database.variables)

# Removing some observations can be done directly using pandas.
#remove = (((database.data.PURPOSE != 1) & (database.data.PURPOSE != 3)) | (database.data.CHOICE == 0))
#database.data.drop(database.data[remove].index,inplace=True)

# Here we use the "biogeme" way for backward compatibility
exclude = (( PURPOSE != 1 ) * (  PURPOSE   !=  3  ) +  ( CHOICE == 0 )) > 0
database.remove(exclude)



ASC_CAR = Beta('ASC_CAR',0,None,None,0)
ASC_TRAIN = Beta('ASC_TRAIN',0,None,None,0)
ASC_SM = Beta('ASC_SM',0,None,None,1)
B_TIME = Beta('B_TIME',0,None,None,0)
B_COST = Beta('B_COST',0,None,None,0)

MU = Beta('MU',1,0,1,0)
MU_EXISTING = Beta('MU_EXISTING',1,1,None,1)
MU_PUBLIC = Beta('MU_PUBLIC',1,1,None,0)
ALPHA_EXISTING = Beta('ALPHA_EXISTING',0.5,0,1,0)
ALPHA_PUBLIC = 1 - ALPHA_EXISTING



SM_COST =  SM_CO   * (  GA   ==  0  ) 
TRAIN_COST =  TRAIN_CO   * (  GA   ==  0  )

TRAIN_TT_SCALED = TRAIN_TT / 100.0
TRAIN_COST_SCALED = DefineVariable('TRAIN_COST_SCALED',\
                                   TRAIN_COST / 100,database)
SM_TT_SCALED = DefineVariable('SM_TT_SCALED', SM_TT / 100.0,database)
SM_COST_SCALED = DefineVariable('SM_COST_SCALED', SM_COST / 100,database)
CAR_TT_SCALED = DefineVariable('CAR_TT_SCALED', CAR_TT / 100,database)
CAR_CO_SCALED = DefineVariable('CAR_CO_SCALED', CAR_CO / 100,database)

V1 = ASC_TRAIN + \
     B_TIME * TRAIN_TT_SCALED + \
     B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + \
     B_TIME * SM_TT_SCALED + \
     B_COST * SM_COST_SCALED
V3 = ASC_CAR + \
     B_TIME * CAR_TT_SCALED + \
     B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1,
     2: V2,
     3: V3}


# Associate the availability conditions with the alternatives
CAR_AV_SP =  DefineVariable('CAR_AV_SP',CAR_AV  * (  SP   !=  0  ),database)
TRAIN_AV_SP =  DefineVariable('TRAIN_AV_SP',TRAIN_AV  * (  SP   !=  0  ),database)

av = {1: TRAIN_AV_SP,
      2: SM_AV,
      3: CAR_AV_SP}

#Definition of nests:
alpha_existing = {1: ALPHA_EXISTING,
                  2:0.0,
                  3:1.0}

alpha_public = {1: ALPHA_PUBLIC,
                2: 1.0,
                3: 0.0}

nest_existing = MU_EXISTING, alpha_existing
nest_public = MU_PUBLIC, alpha_public
nests = nest_existing, nest_public

# The choice model is a cross-nested logit, with availability conditions

logprob = models.logcnlmu(V,av,nests,CHOICE,MU)
prob1 = models.cnlmu(V,av,nests,1,MU)
genelas1 = Derive(prob1,'TRAIN_TT') * TRAIN_TT / prob1
simulate = {'Prob. train': prob1,
            'Elas. 1':genelas1}

class test_11(unittest.TestCase):
    def testEstimationAndSimulation(self):
        biogeme  = bio.BIOGEME(database,logprob)
        biogeme.modelName = "test_11bis"
        results = biogeme.estimate(algorithm=None)
        self.assertAlmostEqual(results.data.logLike,-5214.049,2)
        biosim  = bio.BIOGEME(database,simulate)
        biosim.modelName = "test_11bis_simul"
        simresults = biosim.simulate(results.data.betaValues)
        self.assertAlmostEqual(sum(simresults['Prob. train']),888.3883902853023,1)
        self.assertAlmostEqual(sum(simresults['Elas. 1']),-17896.043955947385,delta=2)
    
        
if __name__ == '__main__':
    unittest.main()






import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.models as models
import unittest
from biogeme.expressions import Beta, DefineVariable, log, bioDraws, MonteCarlo, PanelLikelihoodTrajectory


pandas = pd.read_csv("swissmetro.dat",sep='\t')
database = db.Database("swissmetro",pandas)
database.panel("ID")

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
#print(database.data.describe())

globals().update(database.variables)

# Removing some observations can be done directly using pandas.
#remove = (((database.data.PURPOSE != 1) & (database.data.PURPOSE != 3)) | (database.data.CHOICE == 0))
#database.data.drop(database.data[remove].index,inplace=True)

# Here we use the "biogeme" way for backward compatibility
exclude = (( PURPOSE != 1 ) * (  PURPOSE   !=  3  ) +  ( CHOICE == 0 )) > 0
database.remove(exclude)



ASC_CAR = Beta('ASC_CAR',0.136,None,None,0)
ASC_TRAIN = Beta('ASC_TRAIN',-1,None,None,0)
ASC_SM = Beta('ASC_SM',0,None,None,1)
B_TIME = Beta('B_TIME',-6.3,None,0,0)
B_COST = Beta('B_COST',-3.29,None,0,0)

SIGMA_CAR = Beta('SIGMA_CAR',3.7,None,None,0)
SIGMA_SM = Beta('SIGMA_SM',0.759,None,None,0)
SIGMA_TRAIN = Beta('SIGMA_TRAIN',3.02,None,None,0)

# Define a random parameter, normally distirbuted, designed to be used
# for Monte-Carlo simulation
EC_CAR = SIGMA_CAR * bioDraws('EC_CAR','NORMAL')
EC_SM = SIGMA_SM * bioDraws('EC_SM','NORMAL')
EC_TRAIN = SIGMA_TRAIN * bioDraws('EC_TRAIN','NORMAL')

SM_COST =  SM_CO   * (  GA   ==  0  ) 
TRAIN_COST =  TRAIN_CO   * (  GA   ==  0  )

TRAIN_TT_SCALED = DefineVariable('TRAIN_TT_SCALED',\
                                 TRAIN_TT / 100.0,database)
TRAIN_COST_SCALED = DefineVariable('TRAIN_COST_SCALED',\
                                   TRAIN_COST / 100,database)
SM_TT_SCALED = DefineVariable('SM_TT_SCALED', SM_TT / 100.0,database)
SM_COST_SCALED = DefineVariable('SM_COST_SCALED', SM_COST / 100,database)
CAR_TT_SCALED = DefineVariable('CAR_TT_SCALED', CAR_TT / 100,database)
CAR_CO_SCALED = DefineVariable('CAR_CO_SCALED', CAR_CO / 100,database)

# For latent class 1, whete the time coefficient is zero
V11 = ASC_TRAIN + B_COST * TRAIN_COST_SCALED  + EC_TRAIN
V12 = ASC_SM + B_COST * SM_COST_SCALED + EC_SM
V13 = ASC_CAR + B_COST * CAR_CO_SCALED + EC_CAR

V1 = {1: V11,
      2: V12,
      3: V13}

# For latent class 2, whete the time coefficient is estimated
V21 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED + EC_TRAIN
V22 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED + EC_SM
V23 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED + EC_CAR

V2 = {1: V21,
      2: V22,
      3: V23}


# Associate the availability conditions with the alternatives

CAR_AV_SP =  DefineVariable('CAR_AV_SP',CAR_AV  * (  SP   !=  0  ),database)
TRAIN_AV_SP =  DefineVariable('TRAIN_AV_SP',TRAIN_AV  * (  SP   !=  0  ),database)

av = {1: TRAIN_AV_SP,
      2: SM_AV,
      3: CAR_AV_SP}


# Class membership model
# Class membership model
CLASS_CTE = Beta('CLASS_CTE',0,None,None,0)
CLASS_INC  = Beta('CLASS_INC',0,None,None,0)
W1 = CLASS_CTE + CLASS_INC * INCOME
probClass1 = models.logit({1:W1,2:0},None,1)
probClass2 = models.logit({1:W1,2:0},None,2)

# The choice model is a discrete mixture of logit, with availability conditions
# Conditional to the random variables, likelihood if the individual is
# in class 1
prob1 = PanelLikelihoodTrajectory(models.logit(V1,av,CHOICE))

# The choice model is a discrete mixture of logit, with availability conditions
# Conditional to the random variables, likelihood if the individual is
# in class 2
prob2 = PanelLikelihoodTrajectory(models.logit(V2,av,CHOICE))

# Conditional to the random variables, likelihood for the individual.
probIndiv = probClass1 * prob1 + probClass2 * prob2

# We integrate over the random variables using Monte-Carlo
logprob = log(MonteCarlo(probIndiv))

class test_16(unittest.TestCase):
    def testEstimation(self):
        biogeme = bio.BIOGEME(database,logprob,numberOfDraws=5,seed=10)
        results = biogeme.estimate()
        self.assertAlmostEqual(results.data.logLike,-4102.692581020569,2)

    
if __name__ == '__main__':
    unittest.main()





""" 
Experimental algorithm by Gael Lederrey. Not yet documented.

:author: Michel Bierlaire and Gael Lederrey
:date: Thu Dec 26 17:41:18 2019
"""

import numpy as np
import pandas as pd
import biogeme.optimization as opt
import biogeme.messaging as msg
import biogeme.exceptions as excep
import copy
import time
logger = msg.bioMessage()

class smoothing:

    def __init__(self,windowSize=10):
        self.windowSize = windowSize
        self.f = []
        self.g = []
        self.h = []
        self.batch = []
        self.n = None

    def numberOfValues(self):
        return len(self.f)


    def removeLast(self):
        self.f.pop()
        self.g.pop()
        self.h.pop()
        self.batch.pop()
        
    def add(self, f, g, h, batch, discount = 0.95):
        if g is not None:
            if self.n is None:
                self.n = len(g)
            elif len(g) != self.n:
                raise excep.biogemeError(f'Incompatible dimensions {len(g)} and {self.n}')
        if h is not None:
            if h.shape != (self.n,self.n):
                raise excep.biogemeError(f'Incompatible dimensions {h.shape} and ({self.n},{self.n})')
        if batch <= 0.0 or batch > 1.0:
            raise excep.biogemeError(f'Batch size must be between 0 and 1: {batch}')
        self.f += [f]
        self.g += [g]
        self.h += [h]
        self.batch += [batch]
        return self.f_g_h(discount)

    def f_g_h(self,discount = 0.95):
        if len(self.f) == 0:
            return None, None, None
        if self.batch[-1] == 1.0:
            return self.f[-1], self.g[-1], self.h[-1]

        # The last value has a weight proportional to the batch that
        # was used.  Each other value has a weight proportional to one
        # minus the batch of the last value, times a discounted factor
        scale = [self.batch[-1]] + [discount ** k * (1.0-self.batch[-1]) / (self.windowSize - 1) for k in range(1,min(len(self.f),self.windowSize))]
        normscale = [x / sum(scale) for x in scale]
        f = 0.0
        for k in range(0,min(len(self.f),self.windowSize)):
            f += normscale[k] * self.f[len(self.f)-k-1]

        if self.n is None:
            # No derivative has yet been stored
            return f, None, None
        else:
            g = np.zeros(self.n)
            h = np.zeros(( self.n, self.n))
            gscale = 0.0
            hscale = 0.0
            for k in range(0,min(len(self.f),self.windowSize)):
                i = len(self.f)-k-1
                if self.g[i] is not None:
                    g += scale[k] * self.g[i]
                    gscale += scale[k]
                if self.h[i] is not None:
                    h += scale[k] * self.h[i]
                    hscale += scale[k]

            if hscale == 0.0:
                return f, g / gscale, None
            else:
                return f, g / gscale, h / hscale

def generateCandidateFirstOrder(fct, x, f, g, h, batch, delta, dogleg, maxiter, maxDelta, eta1, eta2):
    for k in range(maxiter):
        if dogleg:
            step,type = opt.dogleg(g,h,delta)
        else: 
            step,type = opt.truncatedConjugateGradient(g,h,delta)
        xc = x + step
        fct.setVariables(xc)
        fc, gc = fct.f_g(batch=batch)
        num = f - fc;
        denom = -np.inner(step,g) - 0.5 * np.inner(step,h @ step)
        rho = num / denom
        if rho < eta1:
            # Failure: reduce the trust region
            delta = np.linalg.norm(step) / 2.0
            success = False
        else:
            success = True
            y = gc - g
            hc = opt.bfgs(h, step, y)
            if rho >= eta2:
                # Enlarge the trust region
                delta = min(2 * delta,maxDelta)
            return success, xc, fc, gc, hc, delta
    return success, None, None, None, None, delta

def generateCandidateSecondOrder(fct, x, f, g, h, batch, delta, dogleg, maxiter, maxDelta, eta1, eta2):
    for k in range(maxiter):
        if dogleg:
            step,type = opt.dogleg(g,h,delta)
        else: 
            step,type = opt.truncatedConjugateGradient(g,h,delta)
        xc = x + step
        fct.setVariables(xc)
        fc, gc, hc = fct.f_g_h(batch=batch)
        num = f - fc;
        denom = -np.inner(step,g) - 0.5 * np.inner(step,h @ step)
        rho = num / denom
        if rho < eta1:
            # Failure: reduce the trust region
            delta = np.linalg.norm(step) / 2.0
            success = False
        else:
            success = True
            if rho >= eta2:
                # Enlarge the trust region
                delta = min(2 * delta,maxDelta)
            return success, xc, fc, gc, hc, delta
    return success, None, None, None, None, delta


        
def hamabs(fct,initBetas,fixedBetas,betaIds,bounds,parameters=None):
    """
    Algorithm inspired by `Lederrey et al. (2019)`
    
    .. _`Lederrey et al. (2019)`: https://transp-or.epfl.ch/documents/technicalReports/LedLurHilBie19.pdf
    
    :param fct: object to calculate the objective function and its derivatives.
    :type obj: optimization.functionToMinimize

    :param initBetas: initial value of the parameters.
    :type initBetas: numpy.array

    :param fixedBetas: betas that stay fixed suring the optimization.
    :type fixedBetas: numpy.array

    :param betaIds: internal identifiers of the non fixed betas.
    :type betaIds: numpy.array

    :param bounds: list of tuples (ell,u) containing the lower and upper bounds for each free parameter. Note that this algorithm does not support bound constraints. Therefore, all the bounds must be None.
    :type bounds: list(tuples)

    :param parameters: dict of parameters to be transmitted to the  optimization routine:             
         - tolerance: when the relative gradient is below that threshold, the algorithm has reached convergence (default:  :math:`\\varepsilon^{\\frac{1}{3}}`);
         - maxiter: the maximum number of iterations (default: 100).

    :type parameters: dict(string:float or int)

    :return: tuple x,nit,nfev,message, where 

            - x is the solution found,
            - nit is the number of iterations performed, 
            - nfev is the number of time that the objective function has been evaluated, and 
            - message is the diagnostic provided by the algorithm.
    :rtype: numpy.array,int,int,string

    """

    for l,u in bounds:
        if l is not None or u is not None:
            raise excep.biogemeError("This algorithm does not handle bound constraints. Remove the bounds, or select another algorithm.")

    tol = np.finfo(np.float64).eps**0.3333
    maxiter = 1000
    # The size of the first batch is such that it can be increased 5 times
    firstBatch = 1.0 / 2.0**4
    # The critical of the batch when BFGS is applied allows for 2 increases
    hybrid = 1.0 / 2.0**2
    firstRadius = 1.0
    # Premature convergence for small batch sizes
    scaleEps = 10.0
    # Maximum number of iterations before updating the batch size
    maxFailure = 2
    
    dogleg = False
    eta1 = 0.01
    eta2 = 0.9

    if parameters is not None:
        if 'tolerance' in parameters:
            tol = parameters['tolerance']
        if 'maxiter' in parameters:
            maxiter = parameters['maxiter']
        if 'firstBatch' in parameters:
            firstBatch = parameters['firstBatch']
        if 'firstRadius' in parameters:
            firstRadius = parameters['firstRadius']
        if 'hybrid' in parameters:
            hybrid = parameters['hybrid']
        if 'maxFailure' in parameters:
            maxFailure = parameters['maxFailure']
        if 'scaleEps' in parameters:
            scaleEps = parameters['scaleEps']
        if 'dogleg' in parameters:
            dogleg = parameters['dogleg']
        if 'eta1' in parameters:
            eta1 = parameters['eta1']
        if 'eta2' in parameters:
            eta2 = parameters['eta2']

    logger.detailed("** Optimization: HAMABS")

    avging = smoothing()

    k = 0
    xk = initBetas
    batch = firstBatch

    fct.setVariables(xk)
    f, g, h = fct.f_g_h(batch=batch)
    avgf, avgg, avgh = avging.add(f, g, h, batch)

    typx = np.ones(np.asarray(xk).shape)
    typf = max(np.abs(f) , 1.0)

    if batch == 1.0:
        relgrad = opt.relativeGradient(xk,f,g,typx,typf)
        if relgrad <= tol:
            message = f"Relative gradient = {relgrad} <= {tol}"
            return xk,0,1,message

    delta = firstRadius
    nfev = 0
    cont = True

    maxDelta = np.finfo(float).max
    minDelta = np.finfo(float).eps

    # Collect statistics per iteration
#    columns = ['Batch','f','relgrad','Time','AbsDiff', 'RelDiff', 'AbsEff', 'RelEff']
#    stats = pd.DataFrame(columns=columns)
    
    while cont:
        start = time.time()
        logger.debug(f'***************** Iteration {k} **************')
        logger.debug(f'N={avging.numberOfValues()} xk={xk} avgf={avgf} delta={delta}')
        k += 1
        if batch <= hybrid:
            success, xc, fc, gc, hc, delta = generateCandidateSecondOrder(fct, xk, avgf, avgg, avgh, batch, delta, dogleg, maxFailure, maxDelta, eta1, eta2)
        else:
            success, xc, fc, gc, hc, delta = generateCandidateFirstOrder(fct, xk, avgf, avgg, avgh, batch, delta, dogleg, maxFailure, maxDelta, eta1, eta2)

        if success:
            xk = xc
            avgf, avgg, avgh = avging.add(fc, gc, hc, batch)
            if batch == 1.0:
                relgrad = opt.relativeGradient(xk,avgf,avgg,typx,typf)
                if relgrad <= tol:
                    message = f"Relative gradient = {relgrad} <= {tol}"
                    cont = False
        else:
            if batch < 1.0:
                batch = min(2.0 * batch, 1.0)
                delta = firstRadius
                if batch <= hybrid:
                    fct.setVariables(xk)
                    f, g, h = fct.f_g_h(batch=batch)
                    avgf, avgg, avgh = avging.add(f, g, h, batch)
                else:
                    fct.setVariables(xk)
                    f, g = fct.f_g(batch=batch)
                    avgf, avgg, _ = avging.add(f, g, None, batch)
                    
        if delta <= minDelta:
            if batch == 1.0:
                message = f"Trust region is too small: {delta}"
                cont = False

        if k == maxiter:
            message = f"Maximum number of iterations reached: {maxiter}"
            cont = False
        logger.detailed(f"{k} f={avgf:10.7g} delta={delta:6.2g} batch={100*batch:6.2g}%")
        elapsed = time.time()-start
#        arow = {'Batch': 100*batch, 'f': avgf, 'relgrad': relgrad, 'Time': elapsed, 'AbsDiff': num, 'RelDiff': relnum, 'AbsEff': num/elapsed, 'RelEff': relnum/elapsed}
#        stats.loc[k] = pd.Series(arow)
#    logger.debug(f'\n{stats.to_string()}')
    logger.detailed(message)
    return xk,k,nfev,message
    
    



#######################################
#
# File: 09normalMixtureAntithetic_halton_1000.py
# Author: Michel Bierlaire, EPFL
# Date: Sat Jul 25 19:14:42 2015
#
#######################################

from biogeme import *
from headers import *
from loglikelihood import *
from statistics import *

#Parameters 
ASC_CAR = 0.137
ASC_TRAIN = -0.402
ASC_SM = 0
B_TIME = -2.26
B_TIME_S = 1.66
B_COST = -1.29

# Define a random parameter, normally distirbuted, designed to be used
# for Monte-Carlo simulation

B_TIME_RND = B_TIME + B_TIME_S * bioDraws('B_TIME_RND')
B_TIME_RND_MINUS = B_TIME - B_TIME_S * bioDraws('B_TIME_RND')

# Utility functions

#If the person has a GA (season ticket) her incremental cost is actually 0 
#rather than the cost value gathered from the
# network data. 
SM_COST =  SM_CO   * (  GA   ==  0  ) 
TRAIN_COST =  TRAIN_CO   * (  GA   ==  0  )

# For numerical reasons, it is good practice to scale the data to
# that the values of the parameters are around 1.0. 
# A previous estimation with the unscaled data has generated
# parameters around -0.01 for both cost and time. Therefore, time and
# cost are multipled my 0.01.

TRAIN_TT_SCALED = DefineVariable('TRAIN_TT_SCALED', TRAIN_TT / 100.0)
TRAIN_COST_SCALED = DefineVariable('TRAIN_COST_SCALED', TRAIN_COST / 100)
SM_TT_SCALED = DefineVariable('SM_TT_SCALED', SM_TT / 100.0)
SM_COST_SCALED = DefineVariable('SM_COST_SCALED', SM_COST / 100)
CAR_TT_SCALED = DefineVariable('CAR_TT_SCALED', CAR_TT / 100)
CAR_CO_SCALED = DefineVariable('CAR_CO_SCALED', CAR_CO / 100)

V1 = ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

V1_MINUS = ASC_TRAIN + B_TIME_RND_MINUS * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2_MINUS = ASC_SM + B_TIME_RND_MINUS * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3_MINUS = ASC_CAR + B_TIME_RND_MINUS * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1,
     2: V2,
     3: V3}

V_MINUS = {1: V1_MINUS,
           2: V2_MINUS,
           3: V3_MINUS}

# Associate the availability conditions with the alternatives

CAR_AV_SP =  DefineVariable('CAR_AV_SP',CAR_AV  * (  SP   !=  0  ))
TRAIN_AV_SP =  DefineVariable('TRAIN_AV_SP',TRAIN_AV  * (  SP   !=  0  ))

av = {1: TRAIN_AV_SP,
      2: SM_AV,
      3: CAR_AV_SP}

# The choice model is a logit, with availability conditions
integrand_plus = bioLogit(V,av,CHOICE)
integrand_minus = bioLogit(V_MINUS,av,CHOICE)
integrand = 0.5 * (integrand_plus + integrand_minus)
simulatedI = MonteCarlo(integrand)

trueI = 0.637849835578 

sampleVariance = \
  MonteCarlo(integrand*integrand) - simulatedI * simulatedI
stderr = (sampleVariance / 200000.0)**0.5
error = simulatedI - trueI

simulate = {'01 Simulated Integral': simulatedI,
            '02 Analytical Integral': trueI,
            '03 Sample variance': sampleVariance,
            '04 Std Error': stderr,
            '05 Error': error}

rowIterator('obsIter') 

BIOGEME_OBJECT.SIMULATE = Enumerate(simulate,'obsIter')

__rowId__ = Variable('__rowId__')
BIOGEME_OBJECT.EXCLUDE = __rowId__ >= 1

#For 2000 normals draws, you need 1000 antithetic.
BIOGEME_OBJECT.PARAMETERS['NbrOfDraws'] = "1000"
BIOGEME_OBJECT.PARAMETERS['RandomDistribution'] = "HALTON"
BIOGEME_OBJECT.DRAWS = { 'B_TIME_RND': 'NORMAL' }



"""Singleton managing the various levels of messages

:author: Michel Bierlaire
:date: Mon Jul 22 16:12:00 2019

"""
import datetime
import biogeme.filenames as bf
import biogeme.version as bv

class Singleton(type):
    """
    A singleton is a class with only one instance
    """
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

class bioMessage(metaclass=Singleton):
    """ Manages the Biogeme messages
    """
    def __init__(self,screenLevel=0):
        """ Constructor
        
        :param screenLevel: level of message that must be displayed on the screen:
        
           - 0: no output (default)
           - 1: warnings only
           - 2: only warnings and general information
           - 3: more verbose
           - 4: debug messages


        :type screenlevel: int
        """
        self.screenLevel = screenLevel
        self.types = {0:'Silent',1:'Warning',2:'General',3:'Detailed',4:'Debug'}
        self.messages = []

    def createLog(self,fileLevel=3,fileName="_biogeme"):
        """ Creates a log file

        :param fileLevel: level of message that must be reported in the file:
        
           - 0: no output
           - 1: warnings only
           - 2: only warnings and general information
           - 3: more verbose (default)
           - 4: debug messages

        :type screenlevel: int

        :param fileName: name of the file (without extension). Default: "_biogeme". A file called _biogeme.log will be created.
        :type fileName: string
        """
        completeFileName = bf.getNewFileName(fileName,"log")
        f = open(completeFileName,"w")
        self.general(f"Log file created: {completeFileName}")
        print(f"*** File created {datetime.datetime.now()} ***",file=f)
        print(f"*** Log file from biogeme {bv.getVersion()} [{bv.versionDate}]",file=f)
        for m,i in self.messages:
            if i <= fileLevel:
                print(m,file=f)
        self.messages = []
        return completeFileName

    def setSilent(self):
        """ Set both screen and file levels to 0 """
        self.screenLevel = 0
        self.fileLevel = 0

    def setWarning(self):
        """ Set both screen and file levels to 1 """
        self.screenLevel = 1
        self.fileLevel = 1
        
    def setGeneral(self):
        """ Set both screen and file levels to 2 """
        self.screenLevel = 2
        self.fileLevel = 2
        
    def setDetailed(self):
        """ Set both screen and file levels to 3 """
        self.screenLevel = 3
        self.fileLevel = 3

    def setDebug(self):
        """ Set both screen and file levels to 4 """
        self.screenLevel = 4
        self.fileLevel = 4
        
    def setScreenLevel(self,level):
        """ Change the level of messaging for the screen

        :param level: level of message that must be displayed on the screen:
        
           - 0: no output
           - 1: warnings only
           - 2: only warnings and general information
           - 3: more verbose
           - 4: debug messages


        :type level: int
        """
        self.screenLevel = level

    def setFileLevel(self,level):
        """ Change the level of messaging for the file

        :param level: level of message that must be reported in the file:
        
           - 0: no output
           - 1: warnings only
           - 2: only warnings and general information
           - 3: more verbose
           - 4: debug messages

        :type level: int
        """
        self.fileLevel = level
        
    def addMessage(self,text,level):
        """ Add a message

        :param text: text of the message.
        :type text: string

        :param level: level of the message

           - 1: warning
           - 2: general information
           - 3: detailed information
           - 4: debug message

        :note: adding a message of level 0 is meaningless, as it correspond to silentmode.
        """
        theLevel = f"< {self.types[level]} >"
        theMessage = f"[{datetime.datetime.now().strftime('%H:%M:%S')}] {theLevel:13} {text}"
        if level != 0:
            self.messages.append((theMessage,level))
            if level <= self.screenLevel:
                print(theMessage)

    def warning(self,text):
        """ Add a warning

        :param text: text of the message.
        :type text: string

        """
        self.addMessage(text,1)

    def general(self,text):
        """ Add a general message

        :param text: text of the message.
        :type text: string

        """
        self.addMessage(text,2)

    def detailed(self,text):
        """ Add a detailed message

        :param text: text of the message.
        :type text: string

        """
        self.addMessage(text,3)

    def debug(self,text):
        """ Add a debug message

        :param text: text of the message.
        :type text: string

        """
        self.addMessage(text,4)
        
    

    



#######################################
#
# File: 01simpleIntegral_mlhs.py
# Author: Michel Bierlaire, EPFL
# Date: Sat Jul 25 11:41:13 2015
#
#######################################
#

from biogeme import *
from headers import *

integrand = exp(bioDraws('U'))
simulatedI = MonteCarlo(integrand)

trueI = exp(1.0) - 1.0 

sampleVariance = \
  MonteCarlo(integrand*integrand) - simulatedI * simulatedI
stderr = (sampleVariance / 200000.0)**0.5
error = simulatedI - trueI

simulate = {'01 Simulated Integral': simulatedI,
            '02 Analytical Integral': trueI,
            '03 Sample variance': sampleVariance,
            '04 Std Error': stderr,
            '05 Error': error}

rowIterator('obsIter') 

BIOGEME_OBJECT.SIMULATE = Enumerate(simulate,'obsIter')

BIOGEME_OBJECT.PARAMETERS['NbrOfDraws'] = "20000"
BIOGEME_OBJECT.PARAMETERS['RandomDistribution'] = "MLHS"

__rowId__ = Variable('__rowId__')
BIOGEME_OBJECT.EXCLUDE = __rowId__ >= 1
BIOGEME_OBJECT.DRAWS = { 'U': 'UNIFORM'}



import numpy as np
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.models as models
from biogeme.expressions import Beta, DefineVariable, bioDraws, MonteCarlo, PanelLikelihoodTrajectory, log
import unittest

pandas = pd.read_csv("swissmetro.dat",sep='\t')
database = db.Database("swissmetro",pandas)

database.panel("ID")

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
#print(database.data.describe())

globals().update(database.variables)

# Removing some observations can be done directly using pandas.
#remove = (((database.data.PURPOSE != 1) & (database.data.PURPOSE != 3)) | (database.data.CHOICE == 0))
#database.data.drop(database.data[remove].index,inplace=True)

# Here we use the "biogeme" way for backward compatibility
exclude = (( PURPOSE != 1 ) * (  PURPOSE   !=  3  ) +  ( CHOICE == 0 )) > 0
database.remove(exclude)

ASC_CAR = Beta('ASC_CAR',0,None,None,0)
ASC_TRAIN = Beta('ASC_TRAIN',0,None,None,0)
ASC_SM = Beta('ASC_SM',0,None,None,1)
B_TIME = Beta('B_TIME',0,None,None,0)
B_COST = Beta('B_COST',0,None,None,0)

SIGMA_CAR = Beta('SIGMA_CAR',0,None,None,0)
SIGMA_SM = Beta('SIGMA_SM',0,None,None,0)
SIGMA_TRAIN = Beta('SIGMA_TRAIN',0,None,None,0)


# Provide my own random number generator to the database.
# See the numpy.random documentation to obtain a list of other distributions.
def theTriangularGenerator(sampleSize, numberOfDraws):
    return np.random.triangular(-1,0,1,(sampleSize,numberOfDraws))

myRandomNumberGenerators = {'TRIANGULAR':(theTriangularGenerator,'Triangulart distribution T(-1,0,1)')}
database.setRandomNumberGenerators(myRandomNumberGenerators)

# Define a random parameter,  with a triangular distribution, designed to be used
# for Monte-Carlo simulation
EC_CAR = SIGMA_CAR * bioDraws('EC_CAR','TRIANGULAR')
EC_SM = SIGMA_SM * bioDraws('EC_SM','TRIANGULAR')
EC_TRAIN = SIGMA_TRAIN * bioDraws('EC_TRAIN','TRIANGULAR')


SM_COST =  SM_CO   * (  GA   ==  0  ) 
TRAIN_COST =  TRAIN_CO   * (  GA   ==  0  )

TRAIN_TT_SCALED = DefineVariable('TRAIN_TT_SCALED',\
                                 TRAIN_TT / 100.0,database)
TRAIN_COST_SCALED = DefineVariable('TRAIN_COST_SCALED',\
                                   TRAIN_COST / 100,database)
SM_TT_SCALED = DefineVariable('SM_TT_SCALED', SM_TT / 100.0,database)
SM_COST_SCALED = DefineVariable('SM_COST_SCALED', SM_COST / 100,database)
CAR_TT_SCALED = DefineVariable('CAR_TT_SCALED', CAR_TT / 100,database)
CAR_CO_SCALED = DefineVariable('CAR_CO_SCALED', CAR_CO / 100,database)

V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED + EC_TRAIN
V2 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED + EC_SM
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED + EC_CAR

# Associate utility functions with the numbering of alternatives
V = {1: V1,
     2: V2,
     3: V3}


# Associate the availability conditions with the alternatives
CAR_AV_SP =  DefineVariable('CAR_AV_SP',CAR_AV  * (  SP   !=  0  ),database)
TRAIN_AV_SP =  DefineVariable('TRAIN_AV_SP',TRAIN_AV  * (  SP   !=  0  ),database)

av = {1: TRAIN_AV_SP,
      2: SM_AV,
      3: CAR_AV_SP}

obsprob = models.logit(V,av,CHOICE)
condprobIndiv = PanelLikelihoodTrajectory(obsprob)
logprob = log(MonteCarlo(condprobIndiv))

class test_26(unittest.TestCase):
    def testEstimation(self):
        biogeme = bio.BIOGEME(database,logprob,seed=10,numberOfDraws=5)
        biogeme.modelName = "test_26"
        results = biogeme.estimate(algorithm=None)
        self.assertAlmostEqual(results.data.logLike,-4577.850124399228,2)

if __name__ == '__main__':
    unittest.main()






"""File 14selectionBias.py

:author: Michel Bierlaire, EPFL
:date: Sun Sep  8 19:26:25 2019

 Example of a nested logit model, with the corrections for endogenous sampling. 
 Three alternatives: Train, Car and Swissmetro
 Train and car are in the same nest.
 SP data
"""
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.models as models
from biogeme.expressions import Beta, DefineVariable

# Read the data
df = pd.read_csv("swissmetro.dat",sep='\t')
database = db.Database("swissmetro",df)

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
#print(database.data.describe())

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Removing some observations can be done directly using pandas.
#remove = (((database.data.PURPOSE != 1) & (database.data.PURPOSE != 3)) | (database.data.CHOICE == 0))
#database.data.drop(database.data[remove].index,inplace=True)

# Here we use the "biogeme" way for backward compatibility
exclude = (( PURPOSE != 1 ) * (  PURPOSE   !=  3  ) +  ( CHOICE == 0 )) > 0
database.remove(exclude)

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR',0,None,None,0)
ASC_TRAIN = Beta('ASC_TRAIN',0,None,None,0)
ASC_SM = Beta('ASC_SM',0,None,None,1)
B_TIME = Beta('B_TIME',0,None,None,0)
B_COST = Beta('B_COST',0,None,None,0)
MU = Beta('MU',2.05,1,None,0)

# Additional parameters designed to estimate the bias due to choice
# based sampling
SB_TRAIN =  Beta( 'SB_TRAIN', 0, None, None, 0)
SB_CAR =  Beta( 'SB_CAR', 0, None, None, 0)
correction = {1: SB_TRAIN,
              2: 0,
              3: SB_CAR}

# Definition of new variables
SM_COST =  SM_CO   * (  GA   ==  0  ) 
TRAIN_COST =  TRAIN_CO   * (  GA   ==  0  )

# Definition of new variables: adding columns to the database 
TRAIN_TT_SCALED = DefineVariable('TRAIN_TT_SCALED',\
                                 TRAIN_TT / 100.0,database)
TRAIN_COST_SCALED = DefineVariable('TRAIN_COST_SCALED',\
                                   TRAIN_COST / 100,database)
SM_TT_SCALED = DefineVariable('SM_TT_SCALED', SM_TT / 100.0,database)
SM_COST_SCALED = DefineVariable('SM_COST_SCALED', SM_COST / 100,database)
CAR_TT_SCALED = DefineVariable('CAR_TT_SCALED', CAR_TT / 100,database)
CAR_CO_SCALED = DefineVariable('CAR_CO_SCALED', CAR_CO / 100,database)

# Definition of the utility functions
V1 = ASC_TRAIN + \
     B_TIME * TRAIN_TT_SCALED + \
     B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + \
     B_TIME * SM_TT_SCALED + \
     B_COST * SM_COST_SCALED
V3 = ASC_CAR + \
     B_TIME * CAR_TT_SCALED + \
     B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1,
     2: V2,
     3: V3}

# Associate the availability conditions with the alternatives
CAR_AV_SP =  DefineVariable('CAR_AV_SP',CAR_AV  * (  SP   !=  0  ),database)
TRAIN_AV_SP =  DefineVariable('TRAIN_AV_SP',TRAIN_AV  * (  SP   !=  0  ),database)
av = {1: TRAIN_AV_SP,
      2: SM_AV,
      3: CAR_AV_SP}

#Definition of nests:
# 1: nests parameter
# 2: list of alternatives
existing = MU , [1,3]
future = 1.0 , [2]
nests = existing,future

# The choice model is a nested logit, with corrections for endogenous sampling
# We first obtain the expression of the Gi function for nested logit.
Gi = models.getMevForNested(V,av,nests)
# Then we calculate the MEV log probability, accountinf for the correction.
logprob = models.logmev_endogenousSampling(V,Gi,av,correction,CHOICE)

# Define level of verbosity
import biogeme.messaging as msg
logger = msg.bioMessage()
logger.setSilent()
#logger.setWarning()
#logger.setGeneral()
#logger.setDetailed()

# Create the Biogeme object
biogeme  = bio.BIOGEME(database,logprob)
biogeme.modelName = "14selectionBias"

# Estimate the parameters
results = biogeme.estimate()
pandasResults = results.getEstimatedParameters()
print(pandasResults)





BIOGEME_OBJECT.PARAMETERS['BTRArmijoBeta1'] = "0.1"
BIOGEME_OBJECT.PARAMETERS['BTRArmijoBeta2'] = "0.9"
BIOGEME_OBJECT.PARAMETERS['BTREta1'] = "0.01"
BIOGEME_OBJECT.PARAMETERS['BTREta2'] = "0.9"
BIOGEME_OBJECT.PARAMETERS['BTRGamma1'] = "0.5"
BIOGEME_OBJECT.PARAMETERS['BTRGamma2'] = "0.5"
BIOGEME_OBJECT.PARAMETERS['BTRIncreaseTRRadius'] = "2"
BIOGEME_OBJECT.PARAMETERS['BTRInitRadius'] = "1"
BIOGEME_OBJECT.PARAMETERS['BTRKappaEpp'] = "0.25"
BIOGEME_OBJECT.PARAMETERS['BTRKappaFrd'] = "0.5"
BIOGEME_OBJECT.PARAMETERS['BTRKappaLbs'] = "0.9"
BIOGEME_OBJECT.PARAMETERS['BTRKappaUbs'] = "0.1"
BIOGEME_OBJECT.PARAMETERS['BTRMaxTRRadius'] = "1e+10"
BIOGEME_OBJECT.PARAMETERS['BTRMinTRRadius'] = "1e-07"
BIOGEME_OBJECT.PARAMETERS['BTRTolerance'] = "6.05545e-06"
BIOGEME_OBJECT.PARAMETERS['BTRTypf'] = "1"
BIOGEME_OBJECT.PARAMETERS['MetropolisHastingsNeighborhoodSize'] = "1"
BIOGEME_OBJECT.PARAMETERS['NormalTruncation'] = "1.96"
BIOGEME_OBJECT.PARAMETERS['TSExpTheta'] = "0.5"
BIOGEME_OBJECT.PARAMETERS['TSFractionGradientRequired'] = "0.1"
BIOGEME_OBJECT.PARAMETERS['TolSchnabelEskow'] = "0.00492157"
BIOGEME_OBJECT.PARAMETERS['cfsqpEps'] = "6.05545e-06"
BIOGEME_OBJECT.PARAMETERS['cfsqpEpsEqn'] = "6.05545e-06"
BIOGEME_OBJECT.PARAMETERS['cfsqpUdelta'] = "0"
BIOGEME_OBJECT.PARAMETERS['missingValue'] = "99999"
BIOGEME_OBJECT.PARAMETERS['sensitivityAnalysisAlpha'] = "0.05"
BIOGEME_OBJECT.PARAMETERS['singularValueThreshold'] = "0.0001"
BIOGEME_OBJECT.PARAMETERS['solvoptErrorArgument'] = "0.0001"
BIOGEME_OBJECT.PARAMETERS['solvoptErrorFunction'] = "1e-06"
BIOGEME_OBJECT.PARAMETERS['tTestThreshold'] = "1.96"
BIOGEME_OBJECT.PARAMETERS['toleranceCheckDerivatives'] = "0.001"
BIOGEME_OBJECT.PARAMETERS['HeaderForWeightInSimulatedResults'] = "__weight"
BIOGEME_OBJECT.PARAMETERS['HeaderOfRowId'] = "__rowId__"
BIOGEME_OBJECT.PARAMETERS['OutputFileForSensitivityAnalysis'] = "__sensitivity.dat"
BIOGEME_OBJECT.PARAMETERS['RandomDistribution'] = "MLHS"
BIOGEME_OBJECT.PARAMETERS['individualId'] = "__rowId__"
BIOGEME_OBJECT.PARAMETERS['optimizationAlgorithm'] = "BIO"
BIOGEME_OBJECT.PARAMETERS['stopFileName'] = "STOP"
BIOGEME_OBJECT.PARAMETERS['warningSign'] = "*"
BIOGEME_OBJECT.PARAMETERS['BTRCheapHessian'] = "1"
BIOGEME_OBJECT.PARAMETERS['BTRExactHessian'] = "1"
BIOGEME_OBJECT.PARAMETERS['BTRIncreaseDraws'] = "2"
BIOGEME_OBJECT.PARAMETERS['BTRInitQuasiNewtonWithBHHH'] = "1"
BIOGEME_OBJECT.PARAMETERS['BTRInitQuasiNewtonWithTrueHessian'] = "1"
BIOGEME_OBJECT.PARAMETERS['BTRMaxIter'] = "1000"
BIOGEME_OBJECT.PARAMETERS['BTRQuasiNewtonUpdate'] = "1"
BIOGEME_OBJECT.PARAMETERS['BTRSignificantDigits'] = "7"
BIOGEME_OBJECT.PARAMETERS['BTRStartDraws'] = "10"
BIOGEME_OBJECT.PARAMETERS['BTRUnfeasibleCGIterations'] = "0"
BIOGEME_OBJECT.PARAMETERS['BTRUsePreconditioner'] = "1"
BIOGEME_OBJECT.PARAMETERS['DumpSensitivityAnalysis'] = "0"
BIOGEME_OBJECT.PARAMETERS['NbrOfDraws'] = "500"
BIOGEME_OBJECT.PARAMETERS['NbrOfDrawsForSensitivityAnalysis'] = "100"
BIOGEME_OBJECT.PARAMETERS['Seed'] = "9021967"
BIOGEME_OBJECT.PARAMETERS['accessFirstDataFromMetaIterator'] = "0"
BIOGEME_OBJECT.PARAMETERS['allowNestedMonteCarlo'] = "0"
BIOGEME_OBJECT.PARAMETERS['biogemeDisplay'] = "0"
BIOGEME_OBJECT.PARAMETERS['biogemeLogDisplay'] = "1"
BIOGEME_OBJECT.PARAMETERS['buildAnalyticalGradient'] = "1"
BIOGEME_OBJECT.PARAMETERS['cfsqpIprint'] = "2"
BIOGEME_OBJECT.PARAMETERS['cfsqpMaxIter'] = "1000"
BIOGEME_OBJECT.PARAMETERS['cfsqpMode'] = "100"
BIOGEME_OBJECT.PARAMETERS['checkDerivatives'] = "0"
BIOGEME_OBJECT.PARAMETERS['computeInitLoglikelihood'] = "1"
BIOGEME_OBJECT.PARAMETERS['decimalPrecisionForSimulation'] = "6"
BIOGEME_OBJECT.PARAMETERS['deriveAnalyticalHessian'] = "1"
BIOGEME_OBJECT.PARAMETERS['dumpDrawsOnFile'] = "0"
BIOGEME_OBJECT.PARAMETERS['firstIdOfLiterals'] = "10000"
BIOGEME_OBJECT.PARAMETERS['includeDataWhenSimulate'] = "0"
BIOGEME_OBJECT.PARAMETERS['maxGcpIter'] = "10000"
BIOGEME_OBJECT.PARAMETERS['maxPrimeNumbers'] = "1000"
BIOGEME_OBJECT.PARAMETERS['monteCarloControlVariateReport'] = "100"
BIOGEME_OBJECT.PARAMETERS['moreRobustToNumericalIssues'] = "0"
BIOGEME_OBJECT.PARAMETERS['numberOfThreads'] = "12"
BIOGEME_OBJECT.PARAMETERS['printPValue'] = "1"
BIOGEME_OBJECT.PARAMETERS['saveUniformDraws'] = "1"
BIOGEME_OBJECT.PARAMETERS['scaleDerivativesInSums'] = "0"
BIOGEME_OBJECT.PARAMETERS['shareOfProcessors'] = "50"
BIOGEME_OBJECT.PARAMETERS['simulateReportForEveryObservation'] = "1"
BIOGEME_OBJECT.PARAMETERS['solvoptDisplay'] = "1"
BIOGEME_OBJECT.PARAMETERS['solvoptMaxIter'] = "1000"
BIOGEME_OBJECT.PARAMETERS['svdMaxIter'] = "150"
BIOGEME_OBJECT.PARAMETERS['useAnalyticalHessianForOptimization'] = "0"
BIOGEME_OBJECT.PARAMETERS['varCovarFromBHHH'] = "0"
BIOGEME_OBJECT.PARAMETERS['warnsForIllegalElements'] = "1"




from biogeme.expressions import *
GROUP=Variable('GROUP')
SURVEY=Variable('SURVEY')
SP=Variable('SP')
ID=Variable('ID')
PURPOSE=Variable('PURPOSE')
FIRST=Variable('FIRST')
TICKET=Variable('TICKET')
WHO=Variable('WHO')
LUGGAGE=Variable('LUGGAGE')
AGE=Variable('AGE')
MALE=Variable('MALE')
INCOME=Variable('INCOME')
GA=Variable('GA')
ORIGIN=Variable('ORIGIN')
DEST=Variable('DEST')
TRAIN_AV=Variable('TRAIN_AV')
CAR_AV=Variable('CAR_AV')
SM_AV=Variable('SM_AV')
TRAIN_TT=Variable('TRAIN_TT')
TRAIN_CO=Variable('TRAIN_CO')
TRAIN_HE=Variable('TRAIN_HE')
SM_TT=Variable('SM_TT')
SM_CO=Variable('SM_CO')
SM_HE=Variable('SM_HE')
SM_SEATS=Variable('SM_SEATS')
CAR_TT=Variable('CAR_TT')
CAR_CO=Variable('CAR_CO')
CHOICE=Variable('CHOICE')



""" Implementation of the pdf and CDF of common distributions

:author:Michel Bierlaire

:data: Thu Apr 23 12:01:49 2015

"""

from biogeme.expressions import *


def normalpdf(x,mu=0.0,s=1.0):
    """
    Normal pdf

    Probability density function of a normal distribution 

    .. math:: f(x;\\mu,\\sigma) = \\frac{1}{\\sigma \\sqrt{2\\pi}} \\exp{-\\frac{(x-\\mu)^2}{2\\sigma^2}}
    
    :param x: location parameter :math:`\\mu` of the Normal distribution. Default: 0.
    :type x: float or biogeme.expression
    :param s: scale parameter :math:`\\sigma` of the Normal distribution. Default: 1.
    :type s: float or biogeme.expression

    :note: It is assumed that :math:`\\sigma > 0`, but it is not verified by the code.

    :return: value of the Normal pdf.
    :rtype: float or biogeme.expression
"""    
    d = -(x-mu)*(x-mu)
    n = 2.0*s*s
    a = d/n
    num = exp(a)
    den = s*2.506628275
    p = num / den
    return p

def lognormalpdf(x,mu,s):
    """
    Log normal pdf

    Probability density function of a log normal distribution 

    .. math:: f(x;\\mu,\\sigma) =  \\frac{1}{x\\sigma \\sqrt{2\\pi}} \\exp{-\\frac{(\\ln x-\\mu)^2}{2\\sigma^2}}

    :param x: location parameter :math:`\\mu` of the lognormal distribution. Default: 0.
    :type x: float or biogeme.expression
    :param s: scale parameter :math:`\\sigma` of the lognormal distribution. Default: 1.
    :type s: float or biogeme.expression

    :note: It is assumed that :math:`\\sigma > 0`, but it is not verified by the code.

    :return: value of the lognormal pdf.
    :rtype: float or biogeme.expression

    """
    d = -(log(x)-mu)*(log(x)-mu)
    n = 2.0*s*s
    a = d/n
    num = exp(a)
    den = x*s*2.506628275
    p = (x>0)* num / den
    return p

def uniformpdf(x,a=-1,b=1.0):
    """
    Uniform pdf

    Probability density function of a uniform distribution.

    .. math::  f(x;a,b) = \\left\\{ \\begin{array}{ll} \\frac{1}{b-a} & \\text{for } x \\in [a,b] \\\\ 0 & \\text{otherwise}\\end{array} \\right.

    :param x: argument of the pdf 
    :type x: float or biogeme.expression
    :param a: lower bound :math:`a` of the distribution. Default: -1.
    :type a: float or biogeme.expression
    :param b: upper bound :math:`b` of the distribution. Default: 1.
    :type b: float or biogeme.expression
    :note: It is assumed that :math:`a < b`, but it is not verified by the code.
    :return: value of the uniform pdf.
    :rtype: float or biogeme.expression
 """
    result = (x < a) * 0.0 + (x >= b) * 0.0 + (x >= a) * (x < b) / (b-a)
    return result

def triangularpdf(x,a=-1.0,b=1.0,c=0.0):
    """
    Triangular pdf

    Probability density function of a triangular distribution

    .. math:: f(x;a,b,c) = \\left\\{ \\begin{array}{ll} 0 & \\text{if } x < a \\\\\\frac{2(x-a)}{(b-a)(c-a)} & \\text{if } a \\leq x < c \\\\\\frac{2(b-x)}{(b-a)(b-c)} & \\text{if } c \\leq x < b \\\\0 & \\text{if } x \\geq b.\\end{array} \\right.

    :param x: argument of the pdf 
    :type x: float or biogeme.expression
    :param a: lower bound :math:`a` of the distribution. Default: -1.
    :type a: float or biogeme.expression
    :param b: upper bound :math:`b` of the distribution. Default: 1.
    :type b: float or biogeme.expression
    :param c: mode :math:`c` of the distribution. Default: 0.
    :type c: float or biogeme.expression
    :note: It is assumed that :math:`a <  c < b`, but it is not verified by the code.
    :return: value of the triangular pdf.
    :rtype: float or biogeme.expression
    
    """
    result = (x < a) * 0.0 + (x >= b) * 0.0 + (x >= a) * (x < c) * 2.0 * ((x-a)/((b-a)*(c-a))) *  (x >= c) * (x < b)  * 2.0 * (b-x) / ((b-a)*(b-c))
    return result

def logisticcdf(x,mu=0.0,s=1.0):
    """
    Logistic CDF

    Cumulative distribution function of a logistic distribution

    .. math:: f(x;\\mu,\\sigma) = \\frac{1}{1+\\exp\\left(-\\frac{x-\\mu}{\\sigma} \\right)} 

    :param x: location parameter :math:`\\mu` of the logistic distribution. Default: 0.
    :type x: float or biogeme.expression
    :param x: scale parameter :math:`\\sigma` of the logistic distribution. Default: 1.
    :type x: float or biogeme.expression
    :note: It is assumed that :math:`\\sigma > 0`, but it is not verified by the code.

    :return: value of the logistic CDF.
    :rtype: float or biogeme.expression

    """
    result = 1.0 /( 1.0 + exp(-(x-mu)/s))
    return result



import pandas as pd
import numpy as np
import biogeme.results as res

structResults = res.bioResults(pickleFile='02oneLatentOrdered.pickle')
structBetas = structResults.getBetaValues()

for k,v in structBetas.items():
    print(f"{k} = Beta('{k}',{v},None,None,0 )")


SIGMA_STAR_Envir02 = Beta('SIGMA_STAR_Envir02',0.7670519646409621,None,None,0 )
SIGMA_STAR_Envir03 = Beta('SIGMA_STAR_Envir03',0.7177819009863683,None,None,0 )
SIGMA_STAR_Mobil11 = Beta('SIGMA_STAR_Mobil11',0.7833224122337571,None,None,0 )
SIGMA_STAR_Mobil14 = Beta('SIGMA_STAR_Mobil14',0.6882826546409856,None,None,0 )
SIGMA_STAR_Mobil16 = Beta('SIGMA_STAR_Mobil16',0.7544205820581054,None,None,0 )
SIGMA_STAR_Mobil17 = Beta('SIGMA_STAR_Mobil17',0.7600628538416172,None,None,0 )
coef_ContIncome_0_4000 = Beta('coef_ContIncome_0_4000',0.08954209471304636,None,None,0 )
coef_ContIncome_10000_more = Beta('coef_ContIncome_10000_more',0.08430692986645968,None,None,0 )
coef_ContIncome_4000_6000 = Beta('coef_ContIncome_4000_6000',-0.2209233080453265,None,None,0 )
coef_ContIncome_6000_8000 = Beta('coef_ContIncome_6000_8000',0.2591889240542216,None,None,0 )
coef_ContIncome_8000_10000 = Beta('coef_ContIncome_8000_10000',-0.5227805784067027,None,None,0 )
delta_1 = Beta('delta_1',0.25196349820243613,None,None,0 )
delta_2 = Beta('delta_2',0.759172317380935,None,None,0 )




"""Implements the function providing names for the output files.

:author: Michel Bierlaire

:date: Tue Mar 26 16:48:40 2019
"""
from pathlib import Path


def getNewFileName(name,ext):
    """
      Generate a file name that does not exist. 

    :param name: name of the file.
    :type name: string
    :param ext: file extension.
    :type ext: string

    :return: name.ext if the file does not exists.  If it does, returns
       name~xx.ext, where xx is the smallest integer such that the
       corresponding file does not exist. It is designed to avoid erasing
       output files inadvertently.
    :rtype: string
    """
    fileName = (name+"."+ext)
    theFile = Path(fileName)
    number = int(0)
    while theFile.is_file():
        fileName = "{}~{:02d}.{}".format(name,number,ext)
        theFile = Path(fileName)
        number += 1
    return fileName



""" Generation of various types of draws.

:author: Michel Bierlaire

:date: Tue Jun 18 19:05:13 2019
"""

import pandas as pd
import numpy as np
import biogeme.exceptions as excep
import biogeme.tools as tools

def getUniform(sampleSize,numberOfDraws,symmetric=False):
    """ Uniform [0,1] or [-1,1] numbers

    :param sampleSize: number of observations for which draws must be
                       generated. If None, a one dimensional array
                       will be generated. If it has a values k, then k
                       series of draws will be generated
    :type sampleSize: int
    :param numberOfDraws: number of draws to generate.
    :type numberOfDraws: int
    :param symmetric: if True, draws from [-1:1] are generated. If False, draws from [0:1] are generated.  Default: False
    :type symmetric: bool
    :return: numpy array with the draws
    :rtype: numpy.array

    Example::

        draws = dr.getUniform(sampleSize=3,numberOfDraws=10,symmetric=False)
        array([[0.13053817, 0.63892308, 0.55031567, 0.26347854, 0.16730932,
                0.77745367, 0.48283887, 0.84247501, 0.20550219, 0.02373537],
               [0.68935846, 0.03363595, 0.36006669, 0.26709364, 0.54907706,
                0.22492104, 0.2494399 , 0.17323209, 0.52370401, 0.54091257],
               [0.40310204, 0.89916711, 0.86065005, 0.94277699, 0.09077065,
                0.40107731, 0.22554722, 0.47693135, 0.14058265, 0.17397031]])

        draws = dr.getUniform(sampleSize=3,numberOfDraws=10,symmetric=True)
        array([[ 0.74403237, -0.27995692,  0.33997421, -0.89405035, -0.129761  ,
                 0.86593325,  0.30657422,  0.82435619,  0.498482  ,  0.24561616],
               [-0.48239607, -0.29257815, -0.98342034,  0.68392813, -0.25379429,
                 0.49359859, -0.26459883,  0.14569724, -0.68860467, -0.40903446],
               [ 0.93251627, -0.85166912,  0.58096917,  0.39289882, -0.65088635,
                 0.40114744, -0.61327161,  0.08900539, -0.20985417,  0.67542226]])
    """
    if numberOfDraws <= 0:
        raise excep.biogemeError(f"Invalid number of draws: {numberOfDraws}.")

    if sampleSize <= 0:
        raise excep.biogemeError(f"Invalid sample size: {sampleSize} when generating draws.")
    totalSize = numberOfDraws * sampleSize
        
    uniformNumbers = np.random.uniform(size=totalSize)
    if symmetric:
        uniformNumbers = 2.0 * uniformNumbers - 1.0

    uniformNumbers.shape = (sampleSize,numberOfDraws)
    return uniformNumbers


def getLatinHypercubeDraws(sampleSize,numberOfDraws,symmetric=False,uniformNumbers=None):
    """ Implementation of the Modified Latin Hypercube Sampling proposed
        by Hess et al, 2006.

    
    :param sampleSize: number of observations for which draws must be
                       generated. If None, a one dimensional array
                       will be generated. If it has a values k, then k
                       series of draws will be generated
    :type sampleSize: int
    :param numberOfDraws: number of draws to generate.
    :type numberOfDraws: int
    :param symmetric: if True, draws from [-1:1] are generated. If False, draws from [0:1] are generated.  Default: False
    :type symmetric: bool
    :param uniformNumbers: numpy with uniformly distributed numbers. If None, the numpy uniform number generator is used.
    :type uniformNumbers: numpy.array

    :return: numpy array with the draws
    :rtype: numpy.array

    Example::

        latinHypercube = dr.getLatinHypercubeDraws(sampleSize=3,numberOfDraws=10)
        array([[0.43362897, 0.5275741 , 0.09215663, 0.94056236, 0.34376868,
                0.87195551, 0.41495219, 0.71736691, 0.23198736, 0.145561  ],
               [0.30520544, 0.78082964, 0.83591146, 0.2733167 , 0.53890906,
                0.61607469, 0.00699715, 0.17179441, 0.7557228 , 0.39733102],
               [0.49676864, 0.67073483, 0.9788854 , 0.5726069 , 0.11894558,
                0.05515471, 0.2640275 , 0.82093696, 0.92034628, 0.64866597]])    
    """
    if numberOfDraws <= 0:
        raise excep.biogemeError(f"Invalid number of draws: {numberOfDraws}.")

    if sampleSize <= 0:
        raise excep.biogemeError(f"Invalid sample size: {sampleSize} when generating draws.")
    totalSize = numberOfDraws * sampleSize
        
    if uniformNumbers is None:
        uniformNumbers = np.random.uniform(size=totalSize)
    else:
        if uniformNumbers.size != totalSize:
            raise excep.biogemeError(f"A total of {totalSize} uniform draws must be provided, and not {uniformNumbers.size}.")

    uniformNumbers.shape = (totalSize,)
    numbers = np.array([(float(i)+ uniformNumbers[i])/float(totalSize) for i in range(totalSize)])
    if symmetric:
        numbers = 2.0 * numbers - 1.0

    np.random.shuffle(numbers)
    numbers.shape = (sampleSize,numberOfDraws)
    return numbers

def getHaltonDraws(sampleSize,numberOfDraws,symmetric=False,base=2,skip=0,shuffled=False):
    """ Generate Halton draws

    :param sampleSize: number of observations for which draws must be
                       generated. If None, a one dimensional array
                       will be generated. If it has a values k, then k
                       series of draws will be generated
    :type sampleSize: int

    :param numberOfDraws: number of draws to generate.
    :type numberOfDraws: int

    :param symmetric: if True, draws from [-1:1] are generated. If False, draws from [0:1] are generated.  Default: False
    :type symmetric: bool

    :param base: generate Halton draws for a given basis. Ideally, it should be a prime number. Default: 2.
    :type base: int

    :param skip: the number of  elements of the sequence to be discarded.
    :type skip: int

    :param shuffled: if True, each series is shuffled
    :type shuffled: bool

    :return: numpy array with the draws
    :rtype: numpy.array

    Example::

        halton = dr.getHaltonDraws(sampleSize=2,numberOfDraws=10,base=3)
        array([[0.33333333, 0.66666667, 0.11111111, 0.44444444, 0.77777778,
                0.22222222, 0.55555556, 0.88888889, 0.03703704, 0.37037037],
               [0.7037037 , 0.14814815, 0.48148148, 0.81481481, 0.25925926,
                0.59259259, 0.92592593, 0.07407407, 0.40740741, 0.74074074]])
    """
    if numberOfDraws <= 0:
        raise excep.biogemeError(f"Invalid number of draws: {numberOfDraws}.")

    if sampleSize <= 0:
        raise excep.biogemeError(f"Invalid sample size: {sampleSize} when generating draws.")
    totalSize = numberOfDraws * sampleSize

    numbers = []
    for i in range(totalSize+1):
        n, denom = 0., 1.
        while i > 0:
            i, remainder = divmod(i, base)
            denom *= base
            n += remainder / denom
        numbers.append(n)
    
    numbers = np.array(numbers[1:])
    if shuffled:
        np.random.shuffle(numbers)

    if symmetric:
        numbers = 2.0 * numbers - 1.0

    numbers.shape = (sampleSize,numberOfDraws)
    return numbers
        
def getAntithetic(unif,sampleSize,numberOfDraws):
    """Returns antithetic uniform draws

    :param unif: function taking two arguments (sampleSize,numberOfDraws) and returning U[0,1] draws
    :type unif: function

    :param sampleSize: number of observations for which draws must be
                       generated. If None, a one dimensional array
                       will be generated. If it has a values k, then k
                       series of draws will be generated
    :type sampleSize: int

    :param numberOfDraws: number of draws to generate.
    :type numberOfDraws: int
    
    :return: numpy array with the antithetic draws
    :rtype: numpy.array

    Example:: 
    
          draws = dr.getAntithetic(dr.getUniform,sampleSize=3,numberOfDraws=10)

          array([[0.48592363, 0.13648133, 0.35925946, 0.32431338, 0.32997936,
                  0.51407637, 0.86351867, 0.64074054, 0.67568662, 0.67002064],
                 [0.89261997, 0.0331808 , 0.30767182, 0.93433648, 0.17196124,
                  0.10738003, 0.9668192 , 0.69232818, 0.06566352, 0.82803876],
                 [0.81095587, 0.96171364, 0.40984817, 0.72177258, 0.16481096,
                  0.18904413, 0.03828636, 0.59015183, 0.27822742, 0.83518904]])

    """
    R = int(numberOfDraws / 2)
    draws = unif(sampleSize,R)
    return np.concatenate((draws,1-draws),axis=1)

def getNormalWichuraDraws(sampleSize,numberOfDraws,uniformNumbers=None,antithetic=False):
    """Generate pseudo-random numbers from a normal distribution N(0,1) 

    It uses the Algorithm AS241 Appl. Statist. (1988) Vol. 37, No. 3,
    which produces the normal deviate z corresponding to a given lower
    tail area of p; z is accurate to about 1 part in :math:`10^{16}`.

    :param sampleSize: number of observations for which draws must be
                       generated. If None, a one dimensional array
                       will be generated. If it has a values k, then k
                       series of draws will be generated
    :type sampleSize: int

    :param numberOfDraws: number of draws to generate.
    :type numberOfDraws: int

    :param uniformNumbers: numpy with uniformly distributed numbers. If None, the numpy uniform number generator is used.
    :type uniformNumbers: numpy.array
    :param antithetic: if True, only half of the draws are actually generated, and the series are completed with their antithetic version.
    :type antithetic: bool

    :return: numpy array with the draws
    :rtype: numpy.array

    Example::

        draws = dr.getNormalWichuraDraws(sampleSize=3,numberOfDraws=10)
        array([[ 0.52418458, -1.04344204, -2.11642482,  0.48257162, -2.67188279,
                -1.89993283,  0.28251041, -0.38424425,  1.53182226,  0.30651874],
               [-0.7937038 , -0.07884121, -0.91005616, -0.98855175,  1.09405753,
                -0.5997651 , -1.70785113,  1.57571384, -0.33208723, -1.03510102],
               [-0.13853654,  0.92595498, -0.80136586,  1.68454196,  0.9955927 ,
                -0.28615154,  2.10635541,  0.0436191 , -0.25417774,  0.01026933]])

    """
    if numberOfDraws <= 0:
        raise excep.biogemeError(f"Invalid number of draws: {numberOfDraws}.")

    if antithetic:
        if 2 * int(numberOfDraws / 2) != numberOfDraws:
            raise excep.biogemeError(f"Please specify an even number of draws for antithetic draws. Requested number of {numberOfDraws}.")
        numberOfDraws = int(numberOfDraws / 2)

    if sampleSize <= 0:
        raise excep.biogemeError(f"Invalid sample size: {sampleSize} when generating draws.")
    totalSize = numberOfDraws * sampleSize

    split1 = 0.425e+00
    split2 = 5.e+00
    const1 = 0.180625e+00
    const2 = 1.6e+00
    a0 = 3.3871328727963666080e+00
    a1 = 1.3314166789178437745e+02
    a2 = 1.9715909503065514427e+03
    a3 = 1.3731693765509461125e+04
    a4 = 4.5921953931549871457e+04
    a5 = 6.7265770927008700853e+04
    a6 = 3.3430575583588128105e+04
    a7 = 2.5090809287301226727e+03
    b1 = 4.2313330701600911252e+01
    b2 = 6.8718700749205790830e+02
    b3 = 5.3941960214247511077e+03
    b4 = 2.1213794301586595867e+04
    b5 = 3.9307895800092710610e+04
    b6 = 2.8729085735721942674e+04
    b7 = 5.2264952788528545610e+03
    c0 = 1.42343711074968357734e+00
    c1 = 4.63033784615654529590e+00
    c2 = 5.76949722146069140550e+00
    c3 = 3.64784832476320460504e+00
    c4 = 1.27045825245236838258e+00
    c5 = 2.41780725177450611770e-01
    c6 = 2.27238449892691845833e-02
    c7 = 7.74545014278341407640e-04
    d1 = 2.05319162663775882187e+00
    d2 = 1.67638483018380384940e+00
    d3 = 6.89767334985100004550e-01
    d4 = 1.48103976427480074590e-01
    d5 = 1.51986665636164571966e-02
    d6 = 5.47593808499534494600e-04
    d7 = 1.05075007164441684324e-09
    e0 = 6.65790464350110377720e+00
    e1 = 5.46378491116411436990e+00
    e2 = 1.78482653991729133580e+00
    e3 = 2.96560571828504891230e-01
    e4 = 2.65321895265761230930e-02
    e5 = 1.24266094738807843860e-03
    e6 = 2.71155556874348757815e-05
    e7 = 2.01033439929228813265e-07
    f1 = 5.99832206555887937690e-01
    f2 = 1.36929880922735805310e-01
    f3 = 1.48753612908506148525e-02
    f4 = 7.86869131145613259100e-04
    f5 = 1.84631831751005468180e-05
    f6 = 1.42151175831644588870e-07
    f7 = 2.04426310338993978564e-15
    
    if uniformNumbers is None:
        uniformNumbers = np.random.uniform(size=totalSize)
    elif uniformNumbers.size != totalSize:
        raise excep.biogemeError(f"A total of {totalSize} uniform draws must be provided, and not {uniformNumbers.size}.")
    uniformNumbers.shape = (totalSize,)

    q = uniformNumbers-0.5
    draws = np.zeros(uniformNumbers.shape)
    r = np.zeros(uniformNumbers.shape)
    cond1 = np.abs(uniformNumbers) <= 0.45 
    r[cond1] = const1 - q[cond1] * q[cond1]
    draws[cond1] = q[cond1] * (((((((a7 * r[cond1] + a6) * r[cond1] + a5) * r[cond1] + a4) * r[cond1] + a3) * r[cond1] + a2) * r[cond1] + a1) * r[cond1] + a0) / (((((((b7 * r[cond1] + b6) * r[cond1] + b5) * r[cond1] + b4) * r[cond1] + b3) * r[cond1] + b2) * r[cond1] + b1) * r[cond1] + 1)
    cond2 = np.abs(uniformNumbers) > 0.45
    cond2a = np.logical_and(cond2,q<0.0)
    cond2b = np.logical_and(cond2,q>=0.0)
    r[cond2a] = uniformNumbers[cond2a]
    r[cond2b] = 1-uniformNumbers[cond2b]
    cond2c = np.logical_and(cond2,r <= 0)
    cond2d = np.logical_and(cond2,r > 0)
    draws[cond2c] = 0.0
    r[cond2d] = np.sqrt(-np.log(r[cond2d]))
    cond2d_a = np.logical_and(cond2d,r <= split2)
    cond2d_b = np.logical_and(cond2d,r > split2)
    r[cond2d_a] = r[cond2d_a] - const2
    draws[cond2d_a] = (((((((c7 * r[cond2d_a] + c6) * r[cond2d_a] + c5) * r[cond2d_a] + c4) * r[cond2d_a] + c3) * r[cond2d_a] + c2) * r[cond2d_a] + c1) * r[cond2d_a] + c0) / (((((((d7 * r[cond2d_a] + d6) * r[cond2d_a] + d5) * r[cond2d_a] + d4) * r[cond2d_a] + d3) * r[cond2d_a] + d2) * r[cond2d_a] + d1) * r[cond2d_a] + 1)
    r[cond2d_b] = r[cond2d_b] - split2
    draws[cond2d_b] = (((((((e7 * r[cond2d_b] + e6) * r[cond2d_b] + e5) * r[cond2d_b] + e4) * r[cond2d_b] + e3) * r[cond2d_b] + e2) * r[cond2d_b] + e1) * r[cond2d_b] + e0) / (((((((f7 * r[cond2d_b] + f6) * r[cond2d_b] + f5) * r[cond2d_b] + f4) * r[cond2d_b] + f3) * r[cond2d_b] + f2) * r[cond2d_b] + f1) * r[cond2d_b] + 1)
    draws[cond2a] = -draws[cond2a]

    draws.shape = (sampleSize,numberOfDraws)

    if antithetic:
        draws = np.concatenate((draws,-draws),axis=1)

    return draws



"""File 06nestedWTP.py

:author: Michel Bierlaire, EPFL
:date: Wed Sep 11 14:01:00 2019

 We use a previously estimated nested logit model.
 Three alternatives: public transporation, car and slow modes.
 RP data.
 We calculate and plot willingness to pay.
"""
import biogeme.messaging as msg
logger = msg.bioMessage()
#logger.setSilent()
#logger.setWarning()
logger.setGeneral()
#logger.setDetailed()


import sys
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.models as models
import biogeme.results as res

from biogeme.expressions import Beta, DefineVariable, Derive

import matplotlib.pyplot as plt

print("Running 06nestedWTP.py...")

# Read the data
df = pd.read_csv("optima.dat",sep='\t')
database = db.Database("optima",df)

confidenceInterval = True

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
#print(database.data.describe())

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Exclude observations such that the chosen alternative is -1
exclude = (Choice == -1.0)
database.remove(exclude)


# Normalize the weights
sumWeight = database.data['Weight'].sum()
normalizedWeight = Weight * 1906 / 0.814484

# Calculate the number of accurences of a value in the database
numberOfMales = database.count("Gender",1)
print(f"Number of males:   {numberOfMales}")
numberOfFemales = database.count("Gender",2)
print(f"Number of females: {numberOfFemales}")
# For more complex conditions, using directly Pandas
unreportedGender = \
                   database.data[(database.data["Gender"] != 1)
                    & (database.data["Gender"] != 2)].count()["Gender"]
print(f"Unreported gender: {unreportedGender}")

# List of parameters. Their value will be set later.
ASC_CAR = Beta('ASC_CAR',0,None,None,0)
ASC_PT = Beta('ASC_PT',0,None,None,1)
ASC_SM = Beta('ASC_SM',0,None,None,0)
BETA_TIME_FULLTIME = Beta('BETA_TIME_FULLTIME',0,None,None,0)
BETA_TIME_OTHER = Beta('BETA_TIME_OTHER',0,None,None,0)
BETA_DIST_MALE = Beta('BETA_DIST_MALE',0,None,None,0)
BETA_DIST_FEMALE = Beta('BETA_DIST_FEMALE',0,None,None,0)
BETA_DIST_UNREPORTED = Beta('BETA_DIST_UNREPORTED',0,None,None,0)
BETA_COST = Beta('BETA_COST',0,None,None,0)

# Definition of variables:
# For numerical reasons, it is good practice to scale the data to
# that the values of the parameters are around 1.0.

# The following statements are designed to preprocess the data.
# It is like creating a new columns in the data file. This
# should be preferred to the statement like
# TimePT_scaled = Time_PT / 200.0
# which will cause the division to be reevaluated again and again,
# throuh the iterations. For models taking a long time to
# estimate, it may make a significant difference.

TimePT_scaled = TimePT / 200
TimeCar_scaled = TimeCar / 200
MarginalCostPT_scaled = MarginalCostPT / 10
CostCarCHF_scaled = CostCarCHF / 10
distance_km_scaled = distance_km / 5

male = (Gender == 1)
female = (Gender == 2)
unreportedGender = (Gender == -1)

fulltime = (OccupStat == 1)
notfulltime = (OccupStat != 1)

# Definition of utility functions:
V_PT = ASC_PT + BETA_TIME_FULLTIME * TimePT_scaled * fulltime + \
       BETA_TIME_OTHER * TimePT_scaled * notfulltime + \
       BETA_COST * MarginalCostPT_scaled
V_CAR = ASC_CAR + \
        BETA_TIME_FULLTIME * TimeCar_scaled * fulltime + \
        BETA_TIME_OTHER * TimeCar_scaled * notfulltime + \
        BETA_COST * CostCarCHF_scaled
V_SM = ASC_SM + \
       BETA_DIST_MALE * distance_km_scaled * male + \
       BETA_DIST_FEMALE * distance_km_scaled * female + \
       BETA_DIST_UNREPORTED * distance_km_scaled * unreportedGender

# Associate utility functions with the numbering of alternatives
V = {0: V_PT,
     1: V_CAR,
     2: V_SM}

# Associate the availability conditions with the alternatives.
# In this example all alternatives are available for each individual.
av = {0: 1,
      1: 1,
      2: 1}

# Definition of the nests:
# 1: nests parameter
# 2: list of alternatives

MU_NOCAR = Beta('MU_NOCAR',1.0,1.0,None,0)

CAR_NEST = 1.0 , [ 1]
NO_CAR_NEST = MU_NOCAR , [ 0, 2]
nests = CAR_NEST, NO_CAR_NEST

WTP_PT_TIME = Derive(V_PT,'TimePT') / Derive(V_PT,'MarginalCostPT')
WTP_CAR_TIME = Derive(V_CAR,'TimeCar') / Derive(V_CAR,'CostCarCHF')

simulate = {'weight': normalizedWeight,
            'WTP PT time': WTP_PT_TIME,
            'WTP CAR time': WTP_CAR_TIME}


biogeme = bio.BIOGEME(database, simulate, removeUnusedVariables=False)
biogeme.modelName = "06nestedWTP"

# Retrieve the values of the parameters.
# First, extract the names of parameters needed for the simulation.
betas = biogeme.freeBetaNames

# Read the estimation results from the file.
results = res.bioResults(pickleFile='01nestedEstimation.pickle')

# Extract the values that are necessary.
betaValues = results.getBetaValues(betas)

# simulatedValues is a Panda dataframe with the same number of rows as
# the database, and as many columns as formulas to simulate.
simulatedValues = biogeme.simulate(betaValues)

wtpcar = (60 * simulatedValues['WTP CAR time'] * simulatedValues['weight']).mean()

# Calculate confidence intervals
b = results.getBetasForSensitivityAnalysis(betas,size=1)

# Returns data frame containing, for each simulated value, the left
# and right bounds of the confidence interval calculated by simulation. 
left,right = biogeme.confidenceIntervals(b,0.9)

wtpcar_left = (60 * left['WTP CAR time'] * left['weight']).mean()
wtpcar_right = (60 * right['WTP CAR time'] * right['weight']).mean()
print(f"Average WTP for car: {wtpcar:.3g} CI:[{wtpcar_left:.3g},{wtpcar_right:.3g}]")


# In this specific case, there are only two distinct values in the
# population: for workers and non workers
print("Unique values: ", [f"{i:.3g}" for i in 60 * simulatedValues['WTP CAR time'].unique()])

# Check the value for groups of the population. Define a function that
# works for any filter to avoid repeating code.
def wtpForSubgroup(filter):
    size = filter.sum()
    sim = simulatedValues[filter]
    totalWeight = sim['weight'].sum()
    weight = sim['weight'] * size / totalWeight
    wtpcar = (60 * sim['WTP CAR time'] * weight ).mean()
    wtpcar_left = (60 * left[filter]['WTP CAR time'] * weight ).mean()
    wtpcar_right = (60 * right[filter]['WTP CAR time'] * weight ).mean()
    return wtpcar, wtpcar_left,wtpcar_right
    
# full time workers. 
filter = database.data['OccupStat'] == 1
w,l,r = wtpForSubgroup(filter)
print(f"WTP car for workers: {w:.3g} CI:[{l:.3g},{r:.3g}]")

# females
filter = database.data['Gender'] == 2
w,l,r = wtpForSubgroup(filter)
print(f"WTP car for females: {w:.3g} CI:[{l:.3g},{r:.3g}]")

# males
filter = database.data['Gender'] == 1
w,l,r = wtpForSubgroup(filter)
print(f"WTP car for males: {w:.3g} CI:[{l:.3g},{r:.3g}]")

# We plot the distribution of WTP in the population. In this case,
# there are only two values

plt.hist(60*simulatedValues['WTP CAR time'],
         weights = simulatedValues['weight'])
plt.xlabel("WTP (CHF/hour)")
plt.ylabel("Individuals")
plt.show()
